import json

from cryptotoolbox.connector import crypto_connector

from scipy import stats
import pandas as pd
from numpy.lib.stride_tricks import as_strided as stride
from cryptotoolbox.realtime import realtime_plotting_utility
from cryptotoolbox.risk_metrics import riskmetrics
from cryptotoolbox.analyzer import market

import plotly.express as px
from itertools import cycle

palette = cycle(px.colors.qualitative.Plotly)

import numpy as np
import talib

def strat_daily_ETH(y, eth='ETH', long_expo = 1., short_expo = 0., rebalancing_day=5):  # On rentre le dataframe
    weekday = y.weekday.values
    cours = y[eth].values
    T = np.size(weekday)
    S = np.zeros(T)
    for t in range(30, T):
        if weekday[t] == rebalancing_day and cours[t - 22] > 0:
            if cours[t] > cours[t - 14] * 1.05 or cours[t] > cours[t - 21] * 1.1:
                S[t] = long_expo
            else:
                S[t] = short_expo
        else:
            S[t] = S[t - 1]
    return S


#def strat_daily_BTC(y, btc='BTC', p1=1.1, p2=1.2, long_expo = 1., short_expo = 0.):  # On rentre le dataframe
def strat_daily_BTC(y, btc='BTC', p1=1.05, p2=1.1, long_expo=1., short_expo=0., rebalancing_day=5):  # On rentre le dataframe
    weekday = y.weekday.values
    cours = y[btc].values
    T = np.size(weekday)
    S = np.zeros(T)
    for t in range(22, T):
        if weekday[t] == rebalancing_day and cours[t - 22] > 0:
            if cours[t] > cours[t - 14] * p1 and cours[t] > cours[t - 21] * p2:
                S[t] = long_expo
            else:
                S[t] = short_expo
        else:
            S[t] = S[t - 1]
    return S

def roll(df, w):
    v = df.values
    d0, d1 = v.shape
    s0, s1 = v.strides
    restricted_length = d0 - (w - 1)
    a = stride(v, (restricted_length, w, d1), (s0, s0, s1))
    rolled_df = pd.concat({
        row: pd.DataFrame(values, columns=df.columns)
        for row, values in zip(df.index[-restricted_length:], a)
    })
    return rolled_df.groupby(level=0)

def strat_weekly(y, btc='BTC', tone=14,ttwo=23,p1=1.01, p2=1.05, long_expo=1., short_expo=0., rebalancing_day=5):  # On rentre le dataframe
    weekday = y.weekday.values
    cours = y[btc].values
    T = np.size(weekday)
    S = np.zeros(T)
    for t in range(ttwo, T):
        if weekday[t] == rebalancing_day and cours[t - ttwo] > 0:
            if cours[t] > cours[t - tone] * p1 and cours[t] > cours[t - ttwo] * p2:
                S[t] = long_expo
            else:
                S[t] = short_expo
        else:
            S[t] = S[t - 1]
    return S

#def strat_daily_BTC(y, btc='BTC', p1=1.1, p2=1.2, long_expo = 1., short_expo = 0.):  # On rentre le dataframe
def strat_daily(y, btc='BTC',tone=15,ttwo=26, p1=1., p2=1.05, long_expo=1., short_expo=0., ):  # On rentre le dataframe
    cours = y[btc].values
    T = np.size(cours)
    S = np.zeros(T)
    for t in range(22, T):
        if cours[t] > cours[t - tone] * p1 and cours[t] > cours[t - ttwo] * p2:
            S[t] = long_expo
        else:
            S[t] = short_expo
    return S

def compute_daily_signal_lo(df=None, core_tokens=['BTC', 'ETH'], extra_tokens=['AVAX'], exclude_core_token=False ,tone=15, ttwo=26, p1btc = 1.,p2btc = 1.05,p1eth = 1.,p2eth = 1.05,lo_sig_suffix='LO',compute_lo_strat=True, plot_html=False, whole_kpis = False, local_root_directory='none'):
    all_tokens = core_tokens + extra_tokens

    for me_token in all_tokens:
        df[f'return_{me_token}'] = df[f'close_{me_token}'].pct_change().fillna(0)
    df.replace([np.inf], 0, inplace=True)
    weekly = pd.DataFrame(index=df.index,
                          data={'SBTC': strat_daily(df, 'close_BTC',tone=tone,ttwo=ttwo, p1=p1btc, p2=p2btc)})
    addw = pd.DataFrame(index=df.index,
                        data={'S{}'.format('ETH'): strat_daily(df, 'close_ETH', tone=tone,ttwo=ttwo ,p1=p1eth, p2=p2eth)})
    weekly = weekly.join(addw, how='left')
    for i in extra_tokens:
        addw = pd.DataFrame(index=df.index, data={'S{}'.format(i): strat_daily(df, f'close_{i}')})
        weekly = weekly.join(addw, how='left')
    dfsig = weekly.copy()
    dfsig2 = dfsig.copy()
    ######
    df = pd.merge(df.copy(), dfsig2.copy(), how='left', right_index=True, left_index=True)
    def curate_signals(row,under='BTC'):
        if abs(row[f'close_{under}']) <= 1e-3:
            return np.nan
        else:
            return row[f'S{ssj}']

    for ssj in core_tokens + extra_tokens:
        go_curate = lambda x: curate_signals(x, under=ssj)
        df[f'S{ssj}'] = df.apply(go_curate, axis=1)

    if exclude_core_token:
        for ssj in core_tokens:
            df[f'S{ssj}'] = 0.

    signal_df = df.copy()
    signal_df = signal_df.fillna(0.)
    signal_df['Counter'] = 0
    for i in all_tokens:
        if 'S{}'.format(i) in signal_df.columns:
            signal_df['Counter'] += signal_df['S{}'.format(i)]

    def expo(y, n_max=3):
        S = np.zeros(np.size(y))
        for i in range(np.size(y)):
            if y[i] > n_max:
                S[i] = 1 / y[i]
            elif y[i] > 0:
                S[i] = 1. / n_max
            else:
                S[i] = 0
        return S

    n_max = int(len(all_tokens) / 2.) + 1
    signal_df['expo'] = expo(signal_df['Counter'].values, n_max=n_max)

    for me_tok in all_tokens:
        if not 'S{}'.format(me_tok) in signal_df.columns:
            signal_df[f'S{me_tok}'] = 0.

    for me_tok in all_tokens:
        signal_df[f'SE{me_tok}'] = signal_df[f'S{me_tok}'] * signal_df['expo']

    for i in all_tokens:
        signal_df[f'S{lo_sig_suffix}{i}'] = signal_df[f'SE{i}'].shift(1)
        signal_df[f'SRAW{lo_sig_suffix}{i}'] = signal_df[f'S{i}'].shift(1)


    signal_df[f'total_expo'] = 0.
    for me_tok in all_tokens:
        signal_df[f'total_expo'] = signal_df[f'total_expo'] + signal_df[f'S{lo_sig_suffix}{me_tok}']

    fin_df = signal_df.copy()
    kpi_df = None
    if compute_lo_strat:
        def compute_return(row, tokens=[]):
            total_return = 0.
            for me_tok in tokens:
                total_return = total_return + row[f'return_{me_tok}'] * row[f'S{lo_sig_suffix}{me_tok}']
            return total_return

        comp_ret = lambda x: compute_return(x, tokens=all_tokens)
        fin_df['TR'] = fin_df.apply(comp_ret, axis=1)
        fin_df['TR'] = fin_df['TR'].fillna(0.)
        me_strat = 'strat'
        fin_df[me_strat] = np.cumprod(1 + fin_df['TR'].values)
        title = 'BTC_daily_lo'
        if plot_html:
            fig = realtime_plotting_utility.plot_multiple_time_series(
                data_df=fin_df[[me_strat]],
                logy=True, split=False,
                put_on_same_scale=False,
                title=title)
            fig.show()


        backtest_df = fin_df[['strat']].copy()

        strat_data_df = backtest_df[['strat']].copy()
        from cryptotoolbox.analyzer import market

        # strat_data_df = strat_data_df[strat_data_df.index >= '2021-02-01'].copy()
        ma = market.MarketAnalyzer(strat_data_df, hourlyze=True)

        print('computing kpis')
        kpi_df = ma.get_kpi().T
        kpi_df = kpi_df.dropna()
        kpi_df = kpi_df[~np.isnan(kpi_df.sharpe)]
        print(kpi_df.head())


        if whole_kpis:
            backtest_df.columns = ['strat']
            riskmetrics.compute_plot_excel_kpis_meterics_metrics(backtest_df, title,
                                                                 local_root_directory,
                                                                 plot_html=True,
                                                                 write_file=True)
    return fin_df.copy(), kpi_df


def compute_daily_signal_rsilo(df=None, core_tokens=['BTC', 'ETH'], rsi_lookback = 8 ,threshold = 60 , lo_sig_suffix='RSILO', compute_lo_strat=True, plot_html=False, local_root_directory='None'):
    assert len(core_tokens) == 2
    first_token = core_tokens[0]
    second_token = core_tokens[1]
    df['hour'] = df.index.hour
    all_dfs = None
    for me_hour in [0, 2, 4, 6, 8, 10, 12, 14, 16, 18, 20, 22]:
        daily_df = df[df['hour'] == me_hour].copy()
        daily_df[f'DAILY_RSI_{first_token}_{me_hour}'] = talib.RSI(daily_df[f'close_{first_token}'], rsi_lookback)
        daily_df[f'DAILY_RSI_{second_token}_{me_hour}'] = talib.RSI(daily_df[f'close_{second_token}'], rsi_lookback)
        if all_dfs is None:
            all_dfs = daily_df.copy()
        else:
            all_dfs = pd.concat([all_dfs.copy(), daily_df.copy()])

    all_dfs = all_dfs.sort_index()
    column_list_btc = []
    column_list_eth = []
    for me_hour in [0, 2, 4, 6, 8, 10, 12, 14, 16, 18, 20, 22]:
        all_dfs[f'DAILY_RSI_{first_token}_{me_hour}'] = all_dfs[f'DAILY_RSI_{first_token}_{me_hour}'].fillna(0.)
        all_dfs[f'DAILY_RSI_{second_token}_{me_hour}'] = all_dfs[f'DAILY_RSI_{second_token}_{me_hour}'].fillna(0.)
        column_list_btc.append(f'DAILY_RSI_{first_token}_{me_hour}')
        column_list_eth.append(f'DAILY_RSI_{second_token}_{me_hour}')

    all_dfs[f'DAILY_RSI_{first_token}'] = all_dfs[column_list_btc].sum(axis=1)
    all_dfs[f'DAILY_RSI_{second_token}'] = all_dfs[column_list_eth].sum(axis=1)

    def compute_dual_RSI(row_df, threshold=60, first_token='BTC', second_token='ETH',lo_sig_suffix='RSILO'):
        RSI_BTC = row_df[f'DAILY_RSI_{first_token}']
        RSI_ETH = row_df[f'DAILY_RSI_{second_token}']
        if RSI_BTC > threshold or RSI_ETH > threshold:
            if RSI_BTC > RSI_ETH:
                return {
                    f'S{lo_sig_suffix}{first_token}': 1,
                    f'S{lo_sig_suffix}{second_token}': 0
                }
            else:
                return {
                    f'S{lo_sig_suffix}{first_token}': 0,
                    f'S{lo_sig_suffix}{second_token}': 1
                }
        else:
            return {
                f'S{lo_sig_suffix}{first_token}': 0,
                f'S{lo_sig_suffix}{second_token}': 0
            }

    go_rsi = lambda x: compute_dual_RSI(x, threshold=threshold, first_token=first_token, second_token=second_token, lo_sig_suffix=lo_sig_suffix)
    all_dfs['signal_gen'] = all_dfs.apply(go_rsi, axis=1)
    all_dfs['signal'] = all_dfs['signal_gen'].shift()
    all_dfs = all_dfs.iloc[1:]

    sig_df = pd.DataFrame().from_records(all_dfs['signal'].values)
    sig_df.index = all_dfs.index

    df[f'return_{first_token}'] = df[f'close_{first_token}'].pct_change()
    df[f'return_{second_token}'] = df[f'close_{second_token}'].pct_change()

    data_df = pd.merge(df.copy(), sig_df.copy(), how='left', left_index=True, right_index=True)
    data_df[f'S{lo_sig_suffix}{first_token}'] = data_df[f'S{lo_sig_suffix}{first_token}'].ffill()
    data_df[f'S{lo_sig_suffix}{second_token}'] = data_df[f'S{lo_sig_suffix}{second_token}'].ffill()

    ### we rebalance monday morning
    def get_filter_date(row):
        ##### to get the proper daily close
        #### we only keep the signal that has been emitted the day after
        if row['hour'] == 0:
            return True
        else:
            return False

    data_df['hour'] = data_df.index.hour
    data_df['rebalancing'] = data_df.apply(get_filter_date, axis=1)

    ##### we rebalance with the close of rebalancing_day :
    def rebalancing_signal_cut(row, token='BTC', lo_sig_suffix='RSILO'):
        if row['rebalancing']:
            return row[f'S{lo_sig_suffix}{token}']
        else:
            return np.nan

    goBTC = lambda x: rebalancing_signal_cut(x, token=first_token,lo_sig_suffix=lo_sig_suffix)
    data_df[f'S{lo_sig_suffix}{first_token}'] = data_df.apply(goBTC, axis=1)
    goETH = lambda x: rebalancing_signal_cut(x, token=second_token,lo_sig_suffix=lo_sig_suffix)
    data_df[f'S{lo_sig_suffix}{second_token}'] = data_df.apply(goETH, axis=1)

    data_df[f'S{lo_sig_suffix}{first_token}'] = data_df[f'S{lo_sig_suffix}{first_token}'].ffill()
    data_df[f'S{lo_sig_suffix}{second_token}'] = data_df[f'S{lo_sig_suffix}{second_token}'].ffill()

    moderation = 1.
    mod_data_df = data_df.copy()

    def compute_dual_return(row, moderation=1., first_token='BTC', second_token='ETH', lo_sig_suffix='RSILO'):
        return row[f'S{lo_sig_suffix}{first_token}'] * row[f'return_{first_token}'] * moderation + row[f'S{lo_sig_suffix}{second_token}'] * row[
            f'return_{second_token}'] * moderation

    go_compute = lambda x: compute_dual_return(x, moderation=moderation, first_token=first_token,second_token=second_token,lo_sig_suffix=lo_sig_suffix)
    mod_data_df['strat_return'] = mod_data_df.apply(go_compute, axis=1)

    mod_data_df['strat_return'] = mod_data_df['strat_return'].fillna(0.)
    me_strat = 'strat'
    mod_data_df[me_strat] = np.cumprod(1 + mod_data_df['strat_return'].values)

    if plot_html:
        fig = realtime_plotting_utility.plot_multiple_time_series(
            data_df=mod_data_df[[me_strat]],
            logy=True, split=False,
            put_on_same_scale=False,
            title=f'Perf{first_token}{second_token}')
        fig.show()

    title = f'DUAL_RSI_{moderation}'
    backtest_df = mod_data_df[[me_strat]].copy()
    backtest_df.columns = ['strat']
    strat_data_df = backtest_df[['strat']].copy()

    daily_df = riskmetrics.filter_daily(strat_data_df)
    daily_df = daily_df.dropna()

    if compute_lo_strat:
        title = 'BTC_daily_rsilo'
        backtest_df = daily_df[['strat']].copy()
        whole_kpis = True
        if whole_kpis:
            strat_data_df = backtest_df[['strat']].copy()
            from cryptotoolbox.analyzer import market

            # strat_data_df = strat_data_df[strat_data_df.index >= '2021-02-01'].copy()
            ma = market.MarketAnalyzer(strat_data_df, hourlyze=True)

            print('computing kpis')
            kpi_df = ma.get_kpi().T
            kpi_df = kpi_df.dropna()
            kpi_df = kpi_df[~np.isnan(kpi_df.sharpe)]
            print(kpi_df.head())

        backtest_df.columns = ['strat']
        riskmetrics.compute_plot_excel_kpis_meterics_metrics(backtest_df, title,
                                                             local_root_directory,
                                                             plot_html=True,
                                                             write_file=True)
    return mod_data_df.copy()

def compute_alpha_signal_optimal_lo(df=None, core_tokens=['BTC', 'ETH'], extra_tokens=['AVAX'],     p1btc = 1.01, p2btc = 1.05, p1eth = 1.01, p2eth = 1.05, tone = 14,ttwo = 23, rebalancing_day=6,lo_sig_suffix='LO',compute_lo_strat=True, plot_html=False, whole_kpis=False,local_root_directory='None'):
    all_tokens = core_tokens + extra_tokens
    df['weekday'] = df.index.weekday
    df['rebalance'] = df.index.weekday == rebalancing_day
    for me_token in all_tokens:
        df[f'return_{me_token}'] = df[f'close_{me_token}'].pct_change().fillna(0)
    df.replace([np.inf], 0, inplace=True)
    weekly = pd.DataFrame(index=df.index,
                          data={'SBTC': strat_weekly(df, 'close_BTC', tone=tone, ttwo=ttwo, p1=p1btc, p2=p2btc, rebalancing_day=rebalancing_day)})
    addw = pd.DataFrame(index=df.index,
                        data={'S{}'.format('ETH'): strat_weekly(df, 'close_ETH',  p1=p1eth,p2=p2eth,rebalancing_day=rebalancing_day)})
    weekly = weekly.join(addw, how='left')
    for i in extra_tokens:
        addw = pd.DataFrame(index=df.index, data={'S{}'.format(i): strat_weekly(df, f'close_{i}', rebalancing_day=rebalancing_day)})
        weekly = weekly.join(addw, how='left')
    dfsig = weekly.copy()
    dfsig2 = dfsig.copy()
    ######
    df = pd.merge(df.copy(), dfsig2.copy(), how='left', right_index=True, left_index=True)
    def curate_signals(row,under='BTC'):
        if abs(row[f'close_{under}']) <= 1e-3:
            return np.nan
        else:
            return row[f'S{ssj}']

    for ssj in core_tokens + extra_tokens:
        go_curate = lambda x: curate_signals(x, under=ssj)
        df[f'S{ssj}'] = df.apply(go_curate, axis=1)

    signal_df = df.copy()
    signal_df = signal_df.fillna(0.)
    signal_df['Counter'] = 0
    for i in all_tokens:
        if 'S{}'.format(i) in signal_df.columns:
            signal_df['Counter'] += signal_df['S{}'.format(i)]

    def expo(y, n_max=3):
        S = np.zeros(np.size(y))
        for i in range(np.size(y)):
            if y[i] > n_max:
                S[i] = 1 / y[i]
            elif y[i] > 0:
                S[i] = 1. / n_max
            else:
                S[i] = 0
        return S

    n_max = int(len(all_tokens) / 2.) + 1
    signal_df['expo'] = expo(signal_df['Counter'].values, n_max=n_max)

    for me_tok in all_tokens:
        if not 'S{}'.format(me_tok) in signal_df.columns:
            signal_df[f'S{me_tok}'] = 0.

    for me_tok in all_tokens:
        signal_df[f'SE{me_tok}'] = signal_df[f'S{me_tok}'] * signal_df['expo']

    for i in all_tokens:
        signal_df[f'S{lo_sig_suffix}{i}'] = signal_df[f'SE{i}'].shift(1)
        signal_df[f'SRAW{lo_sig_suffix}{i}'] = signal_df[f'S{i}'].shift(1)


    signal_df[f'total_expo'] = 0.
    for me_tok in all_tokens:
        signal_df[f'total_expo'] = signal_df[f'total_expo'] + signal_df[f'S{lo_sig_suffix}{me_tok}']

    fin_df = signal_df.copy()
    kpi_df = None
    if compute_lo_strat:
        def compute_return(row, tokens=[]):
            total_return = 0.
            for me_tok in tokens:
                total_return = total_return + row[f'return_{me_tok}'] * row[f'S{lo_sig_suffix}{me_tok}']
            return total_return

        comp_ret = lambda x: compute_return(x, tokens=all_tokens)
        fin_df['TR'] = fin_df.apply(comp_ret, axis=1)
        fin_df['TR'] = fin_df['TR'].fillna(0.)
        me_strat = 'strat'
        fin_df[me_strat] = np.cumprod(1 + fin_df['TR'].values)
        title = 'BTC_weekly_lo'
        if plot_html:
            fig = realtime_plotting_utility.plot_multiple_time_series(
                data_df=fin_df[[me_strat]],
                logy=True, split=False,
                put_on_same_scale=False,
                title=title)
            fig.show()


        backtest_df = fin_df[['strat']].copy()

        strat_data_df = backtest_df[['strat']].copy()
        from cryptotoolbox.analyzer import market

        # strat_data_df = strat_data_df[strat_data_df.index >= '2021-02-01'].copy()
        ma = market.MarketAnalyzer(strat_data_df, hourlyze=True)

        print('computing kpis')
        kpi_df = ma.get_kpi().T
        kpi_df = kpi_df.dropna()
        kpi_df = kpi_df[~np.isnan(kpi_df.sharpe)]
        print(kpi_df.head())


        if whole_kpis:
            backtest_df.columns = ['strat']
            riskmetrics.compute_plot_excel_kpis_meterics_metrics(backtest_df, title,
                                                                 local_root_directory,
                                                                 plot_html=True,
                                                                 write_file=True)
    return fin_df.copy(), kpi_df



def compute_alpha_signal_lo(df=None, core_tokens=['BTC', 'ETH'], extra_tokens=['AVAX'], rebalancing_day=6,lo_sig_suffix='LO',compute_lo_strat=True, plot_html=False):
    all_tokens = core_tokens + extra_tokens
    df['weekday'] = df.index.weekday
    df['rebalance'] = df.index.weekday == rebalancing_day
    for me_token in all_tokens:
        df[f'return_{me_token}'] = df[f'close_{me_token}'].pct_change().fillna(0)
    df.replace([np.inf], 0, inplace=True)
    weekly = pd.DataFrame(index=df.index,
                          data={'SBTC': strat_daily_BTC(df, 'close_BTC', 1, 1, rebalancing_day=rebalancing_day)})
    addw = pd.DataFrame(index=df.index,
                        data={'S{}'.format('ETH'): strat_daily_ETH(df, 'close_ETH', rebalancing_day=rebalancing_day)})
    weekly = weekly.join(addw, how='left')
    for i in extra_tokens:
        addw = pd.DataFrame(index=df.index, data={'S{}'.format(i): strat_daily_BTC(df, f'close_{i}', rebalancing_day=rebalancing_day)})
        weekly = weekly.join(addw, how='left')
    dfsig = weekly.copy()
    dfsig2 = dfsig.copy()
    ######
    df = pd.merge(df.copy(), dfsig2.copy(), how='left', right_index=True, left_index=True)
    def curate_signals(row,under='BTC'):
        if abs(row[f'close_{under}']) <= 1e-3:
            return np.nan
        else:
            return row[f'S{ssj}']

    for ssj in core_tokens + extra_tokens:
        go_curate = lambda x: curate_signals(x, under=ssj)
        df[f'S{ssj}'] = df.apply(go_curate, axis=1)

    signal_df = df.copy()
    signal_df = signal_df.fillna(0.)
    signal_df['Counter'] = 0
    for i in all_tokens:
        if 'S{}'.format(i) in signal_df.columns:
            signal_df['Counter'] += signal_df['S{}'.format(i)]

    def expo(y, n_max=3):
        S = np.zeros(np.size(y))
        for i in range(np.size(y)):
            if y[i] > n_max:
                S[i] = 1 / y[i]
            elif y[i] > 0:
                S[i] = 1. / n_max
            else:
                S[i] = 0
        return S

    n_max = int(len(all_tokens) / 2.) + 1
    signal_df['expo'] = expo(signal_df['Counter'].values, n_max=n_max)

    for me_tok in all_tokens:
        if not 'S{}'.format(me_tok) in signal_df.columns:
            signal_df[f'S{me_tok}'] = 0.

    for me_tok in all_tokens:
        signal_df[f'SE{me_tok}'] = signal_df[f'S{me_tok}'] * signal_df['expo']

    for i in all_tokens:
        signal_df[f'S{lo_sig_suffix}{i}'] = signal_df[f'SE{i}'].shift(1)
        signal_df[f'SRAW{lo_sig_suffix}{i}'] = signal_df[f'S{i}'].shift(1)


    signal_df[f'total_expo'] = 0.
    for me_tok in all_tokens:
        signal_df[f'total_expo'] = signal_df[f'total_expo'] + signal_df[f'S{lo_sig_suffix}{me_tok}']

    fin_df = signal_df.copy()
    if compute_lo_strat:
        def compute_return(row, tokens=[]):
            total_return = 0.
            for me_tok in tokens:
                total_return = total_return + row[f'return_{me_tok}'] * row[f'S{lo_sig_suffix}{me_tok}']
            return total_return

        comp_ret = lambda x: compute_return(x, tokens=all_tokens)
        fin_df['TR'] = fin_df.apply(comp_ret, axis=1)
        fin_df['TR'] = fin_df['TR'].fillna(0.)
        me_strat = 'strat'
        fin_df[me_strat] = np.cumprod(1 + fin_df['TR'].values)
        if plot_html:
            dollar_df = fin_df[['strat']].copy()
            new_names = {'strat': 'Absolute Return Weekly LO Strategy'}
            dollar_df = dollar_df.rename(columns=new_names)
            fig = realtime_plotting_utility.plot_dollar_multiple_time_series(data_df=dollar_df,
                                                                             logy=True, split=False,
                                                                             put_on_same_scale=False,
                                                                             title=f'strategy performance')
            fig.show()

    return fin_df.copy()

def aggregate_alpha_signal(daily_crypto_starting_date=None, daily_crypto_ending_date = None, ssj_against = 'USDT', core_tokens = ['BTC','ETH'], extra_tokens = ['AVAX'] , frequence = 'Frequency.DAILY', local_root_directory=None, refetch_pricing=True):
    aggregated_df = None
    hourly_prices = (frequence == 'Frequency.HOURLY')
    for me_symb in core_tokens + extra_tokens:
        print(f'requesting {me_symb}')
        data_df = None
        try:
            if hourly_prices:
                data_df = crypto_connector.fetch_crypto_hourly_data(ssj=me_symb,
                                                                    ssj_against=ssj_against,
                                                                    local_root_directory=local_root_directory,
                                                                    hourly_return_pkl_filename_suffix='hourly_coin_investigation.pkl',
                                                                    refetch_all=refetch_pricing,
                                                                    daily_crypto_starting_day=daily_crypto_starting_date,
                                                                    daily_crypto_ending_day=daily_crypto_ending_date)
            else:
                data_df = crypto_connector.fetch_crypto_daily_data(ssj=me_symb,
                                                                   ssj_against=ssj_against,
                                                                   local_root_directory=local_root_directory,
                                                                   daily_return_pkl_filename_suffix='daily_coin_investigation.pkl',
                                                                   refetch_all=refetch_pricing,
                                                                   daily_crypto_starting_day=daily_crypto_starting_date,
                                                                   daily_crypto_ending_day=daily_crypto_ending_date,
                                                                   saving=True)
        except Exception as e:
            print(f'trouble {e}')

        data_df = data_df.rename(columns={'close': f'close_{me_symb}'})
        data_df = data_df.rename(columns={'open': f'open_{me_symb}'})
        data_df = data_df.rename(columns={'high': f'high_{me_symb}'})
        data_df = data_df.rename(columns={'low': f'low_{me_symb}'})
        data_df = data_df.rename(columns={'volumeto': f'volume_{me_symb}'})
        data_df = data_df.rename(columns={'volume': f'volume_{me_symb}'})

        data_df = data_df.replace(0, np.nan)
        data_df = data_df[
            [f'close_{me_symb}', f'open_{me_symb}', f'high_{me_symb}', f'low_{me_symb}', f'volume_{me_symb}']]
        if aggregated_df is None:
            aggregated_df = data_df.copy()
        else:
            aggregated_df = pd.merge(aggregated_df.copy(), data_df.copy(), left_index=True, right_index=True)

    return aggregated_df

def compute_alpha_signal_ls(lo_df=None, core_tokens=['BTC', 'ETH'], extra_tokens=['AVAX'], moderating_factor = 0.1, lookback_window =90, pente_window = 30, me_center = 0.5, mix_lo = True, me_short = True,ls_signal_suffix = 'LS',aggregate_and_save = True,plot_subsignal_html = True,extract_subsignal_kpi = True):
    import numpy as np
    import functools
    from cryptotoolbox.signal import signal_utility
    from cryptotoolbox.realtime import realtime_plotting_utility
    from cryptotoolbox.risk_metrics import riskmetrics
    ssjs =  core_tokens + extra_tokens
    univ = '_'.join(ssjs)
    aggregated_signals_df = None
    for ssj in ssjs:
        current = f'univ{univ}_lookback_{lookback_window}_pente_{pente_window}_short_{me_short}_center_{me_center}.plkl'
        print('#########'+current)

        data_df = lo_df.copy()
        print('computing weekly signals')

        weekly_df = data_df.copy()
        weekly_df = weekly_df[weekly_df['rebalance']]

        def compute_ranked_slope(short, center, lagging_df):
            lagging_df['rolling_slope_rank'] = lagging_df['rolling_slope'].rank(pct=True)
            if short :
                lagging_df['rolling_slope_rank_ls'] = 2 * lagging_df['rolling_slope_rank'] - center
                lagging_df['rolling_slope_rank_ls'] = lagging_df['rolling_slope_rank_ls'].clip(-1,1)
            else:
                lagging_df['rolling_slope_rank_ls'] = lagging_df['rolling_slope_rank']

            gen_sig = lagging_df['rolling_slope_rank_ls'].iloc[-1]
            return gen_sig

        def compute_slope(slope_df):
           y = slope_df.values
           slope = stats.linregress(np.arange(len(y)), y).slope
           return slope

        weekly_df['rolling_slope'] = weekly_df[f'close_{ssj}'].rolling(window=pente_window).apply(compute_slope)


        go = functools.partial(compute_ranked_slope,me_short,me_center)
        signal_df = roll(weekly_df, lookback_window).apply(go)

        signal_df = signal_df.to_frame()
        signal_df.columns = ['signal_gen']

        data_df = pd.merge(data_df.copy(), signal_df.copy(),how='left', right_index=True, left_index=True)
        data_df['signal_gen']=data_df['signal_gen'].ffill()

        data_df['signal'] = data_df['signal_gen'].shift()
        if aggregate_and_save:
            sig_to_save = data_df[[f'close_{ssj}','signal']].copy()
            if aggregated_signals_df is None:
                sig_to_save.columns = [f'close_{ssj}',f'S{ls_signal_suffix}{ssj}']
                aggregated_signals_df = sig_to_save.copy()
            else :
                sig_to_save.columns = [f'close_{ssj}',f'S{ls_signal_suffix}{ssj}']
                aggregated_signals_df = pd.merge(aggregated_signals_df.copy(), sig_to_save.copy(), left_index = True, right_index=True)


        data_df = data_df.dropna()
        data_df['epoch_number'] = data_df['rebalance'].cumsum()

        perf_df = data_df.copy()
        perf_df=perf_df.rename(columns={f'close_{ssj}':'close'})
        freqly_df, _ = signal_utility.reconstitute_signal_perf(data=perf_df, transaction_cost=True,
                                                               normalization=False)

        if plot_subsignal_html:
            fig = realtime_plotting_utility.plot_multiple_time_series(freqly_df[['signal']],logy = False,drop_na_inf = False, title = ssj)
            fig.show()
            fig = realtime_plotting_utility.plot_multiple_time_series(freqly_df[['reconstituted_perf', 'reconstituted_under']],put_on_same_scale = False, drop_na_inf = False, title = ssj)
            fig.show()
            fig1 = realtime_plotting_utility.plot_multiple_time_series(freqly_df[['reconstituted_perf', 'reconstituted_under']],logy = False, put_on_same_scale = False, drop_na_inf = False, title = ssj)
            fig1.show()


        if extract_subsignal_kpi:
            def compute_metrics(df, strat=None):
                kpi_df = riskmetrics.get_kpi(df[[strat]])
                return kpi_df.to_dict()[strat]

            me_strat = 'reconstituted_perf'
            go_comp_kpi = lambda x: compute_metrics(x, strat=me_strat)

            epochkpis_df = freqly_df[['epoch_number', me_strat]].groupby(['epoch_number']).apply(go_comp_kpi)
            epochkpis_df = epochkpis_df.sort_index()
            epochkpis_df = pd.DataFrame.from_records(epochkpis_df.to_dict()).T



            strat_data_df = freqly_df[['reconstituted_perf', 'reconstituted_under']].copy()
            #strat_data_df = strat_data_df[strat_data_df.index >= '2021-02-01'].copy()
            ma = market.MarketAnalyzer(strat_data_df, hourlyze = True)

            print(f'computing kpis for underlying {ssj}')
            kpi_df = ma.get_kpi().T
            kpi_df=kpi_df.dropna()
            kpi_df=kpi_df[~np.isnan(kpi_df.sharpe)]
            kpi_df = kpi_df[kpi_df.index == 'reconstituted_perf']
            print(kpi_df.head())

    def curate_signals(row,under='BTC'):
        if abs(row[f'close_{under}']) <= 1e-3:
            return np.nan
        else:
            return row[f'S{ls_signal_suffix}{ssj}']

    for ssj in ssjs:
        go_curate = lambda x: curate_signals(x, under=ssj)
        aggregated_signals_df[f'S{ls_signal_suffix}{ssj}'] = aggregated_signals_df.apply(go_curate, axis=1)

#    for extra_tok in extra_tokens :
#        aggregated_signals_df[f'S{ls_signal_suffix}{extra_tok}'] = aggregated_signals_df[f'S{ls_signal_suffix}{extra_tok}'].fillna(0.)

    aggregated_signals_df = aggregated_signals_df.dropna()
#    aggregated_signals_df = aggregated_signals_df[~aggregated_signals_df[f'S{ls_signal_suffix}BTC'].isna().values]
#    aggregated_signals_df = aggregated_signals_df.fillna(0.)

    if mix_lo:
        aggregated_signals_df = pd.merge(aggregated_signals_df.copy(), lo_df.copy(),suffixes=['','_lo'], how='left',
                                         left_index=True, right_index=True)
        def compute_mix_signals(row_df, core_tokens= None, extra_tokens= None):
            lo_expo = row_df['expo']
            if lo_expo > 0 :
                return {f'SF{tok}':row_df[f'SLO{tok}'] for tok in core_tokens + extra_tokens}
            else :
                return {f'SF{tok}':row_df[f'SLS{tok}'] for tok in core_tokens + extra_tokens}

        go_mix = lambda x : compute_mix_signals(x, core_tokens=core_tokens, extra_tokens=extra_tokens)
        nex_signals_series = aggregated_signals_df.apply(go_mix, axis=1)
        nex_signals_df = pd.DataFrame().from_records(nex_signals_series.values)
        nex_signals_df.index = nex_signals_series.index
        aggregated_signals_df = pd.merge(nex_signals_df.copy(),aggregated_signals_df.copy(),left_index=True, right_index=True)
    else :
        for tok in core_tokens + extra_tokens:
            aggregated_signals_df[f'SF{tok}']=aggregated_signals_df[f'SLS{tok}']

    def compute_ls_expo(row_df, universe = None, moderating_factor = np.nan):
        bullish_ones = []
        bearish_ones = []
        bullish_dic = {}
        bearish_dic = {}
        bullish_magnitude = 0.
        bearish_magnitude = 0.
        compo = {}
        for sig in [f'SF{me_tok}' for me_tok in universe]:
            if row_df[sig] > 0. :
                bullish_ones.append(sig)
                bullish_dic[sig] = row_df[sig]
                bullish_magnitude = bullish_magnitude +row_df[sig]
            if row_df[sig]<0.:
                bearish_ones.append(sig)
                bearish_dic[sig] = row_df[sig]
                bearish_magnitude = bearish_magnitude +row_df[sig]

        for sig in [f'SF{me_tok}' for me_tok in universe]:
            if row_df[sig] > 0. and bullish_magnitude>moderating_factor :
                compo[sig] = row_df[sig]/bullish_magnitude*moderating_factor
                continue
            if row_df[sig] > 0.:
                compo[sig] = row_df[sig]
                continue
            if row_df[sig] < 0. and abs(bearish_magnitude)>moderating_factor:
                compo[sig] = row_df[sig]/abs(bearish_magnitude)*moderating_factor
                continue
            if row_df[sig] < 0.:
                compo[sig] = row_df[sig]
                continue
            compo[sig] = 0.

        final_long_expo = 0.
        final_short_expo = 0.
        for sig in [f'SF{me_tok}' for me_tok in universe]:
            if compo[sig]>0.:
                final_long_expo = final_long_expo + compo[sig]
            if compo[sig]<0.:
                final_short_expo = final_short_expo + compo[sig]

        compo['long_expo']  = final_long_expo
        compo['short_expo'] = final_short_expo
        compo['SCASH'] = 1.-compo['long_expo']+ compo['short_expo']
        return compo

    comp_compo = lambda x : compute_ls_expo(x, universe=ssjs, moderating_factor=moderating_factor)
    aggregated_signals_df['compo'] = aggregated_signals_df.apply(comp_compo, axis = 1)

    normed_sig_df = pd.DataFrame().from_records(aggregated_signals_df['compo'].values)
    normed_sig_df.index = aggregated_signals_df.index

    for me_tok in ssjs:
        aggregated_signals_df['return_{}'.format(me_tok)] = aggregated_signals_df['close_{}'.format(me_tok)].pct_change()
    columns_to_keep = [f'return_{me_tok}' for me_tok in ssjs] + [f'close_{me_tok}' for me_tok in ssjs]
    merged_df =  pd.merge(aggregated_signals_df[columns_to_keep].copy(), normed_sig_df.copy(), left_index = True, right_index = True)
    merged_df = merged_df.iloc[1:]
    merged_df.replace([np.inf, -np.inf], np.nan, inplace=True)
    merged_df = merged_df.fillna(0.)

    print('computing exposition')
    def compute_strat_ret(row_df, universe = None, stable_annulalized_apy=0.03):
        aggregated_return = 0.
        for me_token in universe:
            aggregated_return = aggregated_return + row_df[f'return_{me_token}'] * row_df[f'SF{me_token}']
        daily_rate = (1+stable_annulalized_apy)**(1./365.)-1
        aggregated_return = aggregated_return + row_df[f'SCASH']*daily_rate
        return aggregated_return

    comp_ret = lambda x : compute_strat_ret(x, universe=ssjs)
    merged_df['STRAT_RETURN'] = merged_df.apply(comp_ret, axis=1)

    merged_df['COMBO_STRAT'] = np.cumprod(1 + merged_df['STRAT_RETURN'].values)
    merged_df['DOLLAR_COMBO_STRAT'] = 10000. * np.cumprod(1 + merged_df['STRAT_RETURN'].values)
    return merged_df.copy()

def compute_alpha_metrics(data_df = None, core_tokens=['BTC', 'ETH'], extra_tokens=['AVAX'], moderating_factor = 0.1, frequency='weekly', generate_kpis = True, plot_html=True, stop_loss_quantile = 0.75,take_profit_quantile = 0.75,write_excel_file = True, local_root_directory = None, plot_metrics = True,plot_html_save = True):
    ssjs = core_tokens + extra_tokens
    universe_literal = '_'.join(ssjs)
    if plot_html:
        title = f'HedgeFarm Strategy Universe {universe_literal} moderating factor {moderating_factor}'
        backtest_df = data_df[['COMBO_STRAT']].copy()
        backtest_df.columns = ['strategy']
        fig = realtime_plotting_utility.plot_multiple_time_series(
            data_df=backtest_df,
            logy=True, split=False,
            put_on_same_scale=False,
            title=title)
        fig.show()
    # KPI writing


    if generate_kpis:
        riskmetrics.compute_plot_excel_kpis_meterics_metrics(backtest_df, title,
                                                             local_root_directory, plot_html=True,
                                                             write_file=True)
        btd_kpis, ytd_kpis, past_returns = riskmetrics.compute_extensive_kpis(
            data_df=data_df.copy(), strat='COMBO_STRAT')

        go_comp_kpi = lambda x: compute_metrics(x, strat=me_strat)


        def compute_allocation(row_df, universe=None):
            compo = {}
            for me_token in universe:
                compo[me_token] = row_df[f'SF{me_token}']
            return compo


        comp_alloc = lambda x: compute_allocation(x, universe=ssjs)
        data_df['COMBO_STRAT_compo'] = data_df.apply(comp_alloc, axis=1)

        ############# strategy drawdown
        dd_data_df = data_df.copy() #pd.merge(merged_df.copy(), sol_ohlc_df.copy(), left_index=True,
                              #right_index=True)
        limited_dd_data_df = dd_data_df[dd_data_df.index >= '2021-04-01'].copy()
        limited_dd_data_df = limited_dd_data_df[limited_dd_data_df.index <= '2021-07-01'].copy()

        dd_list = []
        for me_underlying in [f'close_{ssj}' for ssj in ['BTC'] + extra_tokens  ] + ['COMBO_STRAT']:
            dd_display_name = 'drawdown ' + me_underlying.replace('close_', '')
            dd_list.append(dd_display_name)
            ### whole period drawdown
            dd_norm_strat = riskmetrics.drawdown(dd_data_df[me_underlying])
            # merged_df[f'drawdown_{me_underlying}'] = dd_norm_strat

            dd_data_df[dd_display_name] = - dd_norm_strat
            ### limited_period_drawdown
            limited_dd_norm_strat = riskmetrics.drawdown(limited_dd_data_df[me_underlying])
            # merged_df[f'drawdown_{me_underlying}'] = dd_norm_strat
            dd_display_name = 'drawdown ' + me_underlying.replace('close_', '')
            limited_dd_data_df[dd_display_name] = - limited_dd_norm_strat



        if plot_html_save:
            dd_dic = {}
            for me_under in extra_tokens:
                dd_dic[f'drawdown {me_under}'] = me_under


            new_names = {'drawdown BTC': 'Bitcoin', 'drawdown COMBO_STRAT': 'Absolute Return DeFi Strategy'}

            new_names.update(dd_dic)

            dd_data_df = dd_data_df.rename(columns=new_names)
            fig = realtime_plotting_utility.plot_multiple_drawdowns(dd_data_df,
                                                                    list(new_names.values()))
            fig.show()

            limited_dd_data_df = limited_dd_data_df.rename(columns=new_names)
            fig = realtime_plotting_utility.plot_multiple_drawdowns(limited_dd_data_df,
                                                                    list(new_names.values()))
            fig.show()

            dollar_df = data_df[['DOLLAR_COMBO_STRAT']].copy()
            new_names = {'DOLLAR_COMBO_STRAT': 'Absolute Return DeFi Strategy'}
            dollar_df = dollar_df.rename(columns=new_names)
            fig = realtime_plotting_utility.plot_dollar_multiple_time_series(data_df=dollar_df,
                                                                             logy=True, split=False,
                                                                             put_on_same_scale=False,
                                                                             title=f'performance (exposition moderating factor {moderating_factor}  rebalancing {frequency})')
            fig.show()

            fig = realtime_plotting_utility.plot_multiple_time_series(
                data_df=data_df[['long_expo', 'short_expo']],
                logy=False, split=False,
                put_on_same_scale=False,
                title=f'long/short exposition (exposition moderating factor {moderating_factor}  rebalancing {frequency})')
            fig.show()

        print('computing KPIs per epoch')

        me_strat = 'COMBO_STRAT'
        whole_period_kpi_df = riskmetrics.get_kpi(data_df[[me_strat]])
        whole_period_kpi_df = whole_period_kpi_df.iloc[1:]

        kpi_df = riskmetrics.get_kpi(data_df[[me_strat]])
        stub = f'mf_{moderating_factor}_strat_{me_strat}'
        kpi_df.columns = [stub]
        to_store_df = data_df.copy()
        to_store_df['Date'] = to_store_df.index

        #### new fresh weights on monday
        to_store_df['has_just_rebalanced'] = to_store_df.index.weekday == 0

        to_store_df['epoch_number'] = to_store_df['has_just_rebalanced'].cumsum()

        epoch_df = to_store_df[to_store_df['has_just_rebalanced']].copy()
        epoch_df = epoch_df[
            ['epoch_number', me_strat, 'long_expo', 'short_expo', f'{me_strat}_compo']]


        def compute_metrics(df, strat=None):
            kpi_df = riskmetrics.get_kpi(df[[strat]])
            return kpi_df.to_dict()[strat]


        go_comp_kpi = lambda x: compute_metrics(x, strat=me_strat)

        epochkpis_df = to_store_df[['epoch_number', me_strat]].groupby(['epoch_number']).apply(
            go_comp_kpi)
        epochkpis_df = epochkpis_df.sort_index()
        epochkpis_df = pd.DataFrame.from_records(epochkpis_df.to_dict()).T
        final_df = pd.merge(epoch_df.copy(), epochkpis_df.copy(), left_on='epoch_number',
                            right_index=True)
        final_df = final_df.drop(columns=[me_strat])
        final_df = final_df.rename(columns={f'{me_strat}_compo': 'COMPO'})

        final_df['STRAT'] = me_strat
        final_df['STRATPARAMS'] = f'{me_strat}_{moderating_factor}'
        # final_df['frequency'] = frequency
        # final_df['moderation'] = moderating_factor
        # if moderating_factor >0.5:
        #     final_df['subscription_asset'] = me_strat.replace('STRAT','').replace('_','')
        # else:
        #     final_df['subscription_asset'] = 'USDC'
        final_df['subscription_asset'] = 'USDC'
        final_df['date'] = final_df.index

        freq_aggregated_df = final_df.copy()
        freq_aggregated_df = freq_aggregated_df.sort_values(by='epoch_number')
        # freq_aggregated_df.index = freq_aggregated_df.index.tz_localize(None)

        av_dd_df = freq_aggregated_df.groupby(['STRATPARAMS'])['mdd'].quantile(q=stop_loss_quantile)
        qt_gain_df = freq_aggregated_df.groupby(['STRATPARAMS'])['simple_return'].quantile(
            q=take_profit_quantile)

        av_dd_df = av_dd_df.to_frame()
        qt_gain_df = qt_gain_df.to_frame()

        av_dd_df.columns = ['stop loss']
        qt_gain_df.columns = ['take profit']
        av_dd_df = av_dd_df.round(decimals=2)
        qt_gain_df = qt_gain_df.round(decimals=2)

        from sklearn.preprocessing import KBinsDiscretizer


        def categorize(data_df=None, column_to_categorize=None):
            est = KBinsDiscretizer(n_bins=3, encode='ordinal', strategy='uniform')
            est.fit(abs(data_df[column_to_categorize].values.reshape(-1, 1)))
            binarized_data = est.transform(abs(data_df[column_to_categorize].values.reshape(-1, 1)))
            data_df[column_to_categorize + '_rank'] = binarized_data
            return data_df.copy()


        av_dd_df_categorized = categorize(data_df=av_dd_df, column_to_categorize='stop loss')

        freq_aggregated_df = pd.merge(freq_aggregated_df.copy(), av_dd_df.copy(),
                                      left_on='STRATPARAMS', right_index=True)
        freq_aggregated_df = pd.merge(freq_aggregated_df.copy(), qt_gain_df.copy(),
                                      left_on='STRATPARAMS', right_index=True)

        ######## computing metrics around strats and epochs
        freq_aggregated_df = freq_aggregated_df.rename(columns={
            'date': 'date',
            'epoch_number': 'epoch',
            'long_expo': 'long risk allocation',
            'short_expo': 'short risk allocation',
            'COMPO': 'composition',
            'simple_return': 'epoch realized return',
            'annual_return': 'epoch realized annualized return',
            'annual_volatility': 'epoch realized annualized volatility',
            'sharpe': 'epoch realized sharpe ratio',
            'calmar': 'epoch realized calmar ratio',
            'mdd': 'epoch realized max drawdown',
            'STRAT': 'strategy',
            'STRATPARAMS': 'strategy parameters',
            'subscription_asset': 'subscription asset'})
        freq_aggregated_df = freq_aggregated_df[
            ['date', 'strategy', 'strategy parameters', 'subscription asset', 'composition',
             'epoch', 'long risk allocation', 'short risk allocation', 'stop loss', 'take profit',
             'epoch realized return', 'epoch realized annualized return',
             'epoch realized annualized volatility',
             'epoch realized sharpe ratio', 'epoch realized calmar ratio',
             'epoch realized max drawdown']]

        freq_aggregated_df = freq_aggregated_df.sort_values(by='epoch')


        new_epoch_details = freq_aggregated_df.iloc[-1].copy()
        new_epoch_details_dict = new_epoch_details.to_dict()
        new_epoch_details_dict['date'] = str(new_epoch_details_dict['date'])

        compo_dic = new_epoch_details_dict['composition']
        clean_compo = {}
        for key, value in compo_dic.items():
            clean_compo[key] = f'{round(value * 100, 2)}%'

        new_epoch_details_dict['composition'] = clean_compo

        payload_message = json.dumps(new_epoch_details_dict)

        payload_message = payload_message.replace('"', '')

        date = new_epoch_details_dict['date']
        strat = new_epoch_details_dict['strategy']
        params = new_epoch_details_dict['strategy parameters']
        sub_ass = new_epoch_details_dict['subscription asset']
        composition = new_epoch_details_dict['composition']
        epoch = new_epoch_details_dict['epoch']
        lr = new_epoch_details_dict['long risk allocation']
        sr = new_epoch_details_dict['short risk allocation']
        sl = new_epoch_details_dict['stop loss']
        tp = new_epoch_details_dict['take profit']


        if plot_metrics:
            to_plot_df = freq_aggregated_df[
                ['epoch realized return', 'epoch realized max drawdown']].copy()
            fig = realtime_plotting_utility.plot_multiple_bar_series(
                data_df=to_plot_df[['epoch realized return']] * 100.,
                logy=False, split=False,
                put_on_same_scale=False,
                title=f'epoch realized return in percent (exposition moderating factor {moderating_factor}  epoch frequency {frequency})')
            fig.show()
            fig = realtime_plotting_utility.plot_multiple_bar_series(
                data_df=to_plot_df[['epoch realized max drawdown']] * 100.,
                logy=False, split=False,
                put_on_same_scale=False,
                title=f'epoch realized max drawdown in percent (exposition moderating factor {moderating_factor}  epoch frequency {frequency})')
            fig.show()
        freq_aggregated_df['year'] = freq_aggregated_df.index.year
        freq_aggregated_df['month'] = freq_aggregated_df.index.month


        def rename_month(row):
            if row['month'] == 1:
                return 'January'
            if row['month'] == 2:
                return 'February'
            if row['month'] == 3:
                return 'March'
            if row['month'] == 4:
                return 'April'
            if row['month'] == 5:
                return 'May'
            if row['month'] == 6:
                return 'June'
            if row['month'] == 7:
                return 'July'
            if row['month'] == 8:
                return 'August'
            if row['month'] == 9:
                return 'September'
            if row['month'] == 10:
                return 'October'
            if row['month'] == 11:
                return 'November'
            if row['month'] == 12:
                return 'December'

        freq_aggregated_df['month'] = freq_aggregated_df.apply(rename_month, axis=1)
        freq_aggregated_df = freq_aggregated_df.set_index(['year', 'month'])
        freq_aggregated_df = freq_aggregated_df.drop(columns=['date', 'epoch'])



        if write_excel_file:
            writer = pd.ExcelWriter(local_root_directory + f'{universe_literal}_weekly_ls_epochs_details_ls.xlsx',
                                    engine='xlsxwriter')
            freq_aggregated_df.to_excel(writer, sheet_name='epoch')
            writer.save()

            writer = pd.ExcelWriter(
                local_root_directory + f'{universe_literal}_kpi_df_ls_strategy.xlsx',
                engine='xlsxwriter')
            # whole_period_kpi_df.to_excel(writer, sheet_name=frequency)
            btd_kpis = btd_kpis.to_frame()
            btd_kpis.columns = [me_strat]
            past_returns = past_returns.to_frame()
            past_returns.columns = [me_strat]

            btd_kpis.to_excel(writer, sheet_name='advanced_btd')
            ytd_kpis.to_excel(writer, sheet_name='advanced_ytd')
            past_returns.to_excel(writer, sheet_name='common')
            writer.save()

    return payload_message




def compute_alpha_signal_rsilo(df=None, core_tokens=['BTC', 'ETH'], rsi_lookback = 8 ,threshold = 60 , rebalancing_day=6,lo_sig_suffix='RSILO', compute_lo_strat=True, plot_html=False, local_root_directory='None'):
    assert len(core_tokens) == 2
    first_token = core_tokens[0]
    second_token = core_tokens[1]
    df['hour'] = df.index.hour
    all_dfs = None
    for me_hour in [0, 2, 4, 6, 8, 10, 12, 14, 16, 18, 20, 22]:
        daily_df = df[df['hour'] == me_hour].copy()
        daily_df[f'DAILY_RSI_{first_token}_{me_hour}'] = talib.RSI(daily_df[f'close_{first_token}'], rsi_lookback)
        daily_df[f'DAILY_RSI_{second_token}_{me_hour}'] = talib.RSI(daily_df[f'close_{second_token}'], rsi_lookback)
        if all_dfs is None:
            all_dfs = daily_df.copy()
        else:
            all_dfs = pd.concat([all_dfs.copy(), daily_df.copy()])

    all_dfs = all_dfs.sort_index()
    column_list_btc = []
    column_list_eth = []
    for me_hour in [0, 2, 4, 6, 8, 10, 12, 14, 16, 18, 20, 22]:
        all_dfs[f'DAILY_RSI_{first_token}_{me_hour}'] = all_dfs[f'DAILY_RSI_{first_token}_{me_hour}'].fillna(0.)
        all_dfs[f'DAILY_RSI_{second_token}_{me_hour}'] = all_dfs[f'DAILY_RSI_{second_token}_{me_hour}'].fillna(0.)
        column_list_btc.append(f'DAILY_RSI_{first_token}_{me_hour}')
        column_list_eth.append(f'DAILY_RSI_{second_token}_{me_hour}')

    all_dfs[f'DAILY_RSI_{first_token}'] = all_dfs[column_list_btc].sum(axis=1)
    all_dfs[f'DAILY_RSI_{second_token}'] = all_dfs[column_list_eth].sum(axis=1)

    def compute_dual_RSI(row_df, threshold=60, first_token='BTC', second_token='ETH',lo_sig_suffix='RSILO'):
        RSI_BTC = row_df[f'DAILY_RSI_{first_token}']
        RSI_ETH = row_df[f'DAILY_RSI_{second_token}']
        if RSI_BTC > threshold or RSI_ETH > threshold:
            if RSI_BTC > RSI_ETH:
                return {
                    f'S{lo_sig_suffix}{first_token}': 1,
                    f'S{lo_sig_suffix}{second_token}': 0
                }
            else:
                return {
                    f'S{lo_sig_suffix}{first_token}': 0,
                    f'S{lo_sig_suffix}{second_token}': 1
                }
        else:
            return {
                f'S{lo_sig_suffix}{first_token}': 0,
                f'S{lo_sig_suffix}{second_token}': 0
            }

    go_rsi = lambda x: compute_dual_RSI(x, threshold=threshold, first_token=first_token, second_token=second_token, lo_sig_suffix=lo_sig_suffix)
    all_dfs['signal_gen'] = all_dfs.apply(go_rsi, axis=1)
    all_dfs['signal'] = all_dfs['signal_gen'].shift()
    all_dfs = all_dfs.iloc[1:]

    sig_df = pd.DataFrame().from_records(all_dfs['signal'].values)
    sig_df.index = all_dfs.index

    df[f'return_{first_token}'] = df[f'close_{first_token}'].pct_change()
    df[f'return_{second_token}'] = df[f'close_{second_token}'].pct_change()

    data_df = pd.merge(df.copy(), sig_df.copy(), how='left', left_index=True, right_index=True)
    data_df[f'S{lo_sig_suffix}{first_token}'] = data_df[f'S{lo_sig_suffix}{first_token}'].ffill()
    data_df[f'S{lo_sig_suffix}{second_token}'] = data_df[f'S{lo_sig_suffix}{second_token}'].ffill()

    ### we rebalance monday morning
    def get_filter_date(row,rebalancing_day=6):
        ##### to get the proper daily close
        #### we only keep the signal that has been emitted the day after
        next_day = (rebalancing_day+1)%7
        if row['weekday'] == next_day and row['hour'] == 0:
            return True
        else:
            return False

    data_df['weekday'] = data_df.index.weekday
    data_df['hour'] = data_df.index.hour
    go_reb = lambda x : get_filter_date(x, rebalancing_day=rebalancing_day)
    data_df['rebalancing'] = data_df.apply(go_reb, axis=1)
    ##### we rebalance with the close of rebalancing_day :
    def rebalancing_signal_cut(row, token='BTC', lo_sig_suffix='RSILO'):
        if row['rebalancing']:
            return row[f'S{lo_sig_suffix}{token}']
        else:
            return np.nan

    goBTC = lambda x: rebalancing_signal_cut(x, token=first_token,lo_sig_suffix=lo_sig_suffix)
    data_df[f'S{lo_sig_suffix}{first_token}'] = data_df.apply(goBTC, axis=1)
    goETH = lambda x: rebalancing_signal_cut(x, token=second_token,lo_sig_suffix=lo_sig_suffix)
    data_df[f'S{lo_sig_suffix}{second_token}'] = data_df.apply(goETH, axis=1)

    data_df[f'S{lo_sig_suffix}{first_token}'] = data_df[f'S{lo_sig_suffix}{first_token}'].ffill()
    data_df[f'S{lo_sig_suffix}{second_token}'] = data_df[f'S{lo_sig_suffix}{second_token}'].ffill()

    moderation = 1.
    mod_data_df = data_df.copy()

    def compute_dual_return(row, moderation=1., first_token='BTC', second_token='ETH', lo_sig_suffix='RSILO'):
        return row[f'S{lo_sig_suffix}{first_token}'] * row[f'return_{first_token}'] * moderation + row[f'S{lo_sig_suffix}{second_token}'] * row[
            f'return_{second_token}'] * moderation

    go_compute = lambda x: compute_dual_return(x, moderation=moderation, first_token=first_token,second_token=second_token,lo_sig_suffix=lo_sig_suffix)
    mod_data_df['strat_return'] = mod_data_df.apply(go_compute, axis=1)

    mod_data_df['strat_return'] = mod_data_df['strat_return'].fillna(0.)
    me_strat = 'strat'
    mod_data_df[me_strat] = np.cumprod(1 + mod_data_df['strat_return'].values)

    if plot_html:
        fig = realtime_plotting_utility.plot_multiple_time_series(
            data_df=mod_data_df[[me_strat]],
            logy=True, split=False,
            put_on_same_scale=False,
            title=f'Perf{first_token}{second_token}')
        fig.show()

    title = f'DUAL_RSI_{moderation}'
    backtest_df = mod_data_df[[me_strat]].copy()
    backtest_df.columns = ['strat']
    strat_data_df = backtest_df[['strat']].copy()

    daily_df = riskmetrics.filter_daily(strat_data_df)
    daily_df = daily_df.dropna()

    if compute_lo_strat:
        title = 'BTC_weekly_rsilo'
        backtest_df = daily_df[['strat']].copy()
        whole_kpis = True
        if whole_kpis:
            strat_data_df = backtest_df[['strat']].copy()
            from cryptotoolbox.analyzer import market

            # strat_data_df = strat_data_df[strat_data_df.index >= '2021-02-01'].copy()
            ma = market.MarketAnalyzer(strat_data_df, hourlyze=True)

            print('computing kpis')
            kpi_df = ma.get_kpi().T
            kpi_df = kpi_df.dropna()
            kpi_df = kpi_df[~np.isnan(kpi_df.sharpe)]
            print(kpi_df.head())

        backtest_df.columns = ['strat']
        riskmetrics.compute_plot_excel_kpis_meterics_metrics(backtest_df, title,
                                                             local_root_directory,
                                                             plot_html=True,
                                                             write_file=True)
    return mod_data_df.copy()


def compute_alpha_signal_dynamical_ls_neutral(lo_df=None, core_tokens=['BTC', 'ETH'], extra_tokens=['AVAX'], moderating_factor = 0.1, lookback_window =5, pente_window = 10, me_center = 0.7, mix_lo = True, me_short = True,ls_signal_suffix = 'LS',aggregate_and_save = True,plot_subsignal_html = True,extract_subsignal_kpi = True):
    import numpy as np
    import functools
    from cryptotoolbox.signal import signal_utility
    from cryptotoolbox.realtime import realtime_plotting_utility
    from cryptotoolbox.risk_metrics import riskmetrics
    ssjs =  core_tokens + extra_tokens
    univ = '_'.join(ssjs)
    aggregated_signals_df = None

    for ssj in ssjs:
        current = f'univ{univ}_lookback_{lookback_window}_pente_{pente_window}_short_{me_short}_center_{me_center}.plkl'
        print('#########'+current)

        data_df = lo_df.copy()
        print('computing weekly signals')

        weekly_df = data_df.copy()
        weekly_df = weekly_df[weekly_df['rebalance']]

        def is_growing(a_np):
            return np.all(a_np[:-1] <= a_np[1:])

        def is_decreasing(a_np):
            return np.all(a_np[:-1] >= a_np[1:])

        def compute_ranked_slope(short, center, lagging_df):
            is_growingg = is_growing(lagging_df['rolling_slope'].values)
            is_decreasingg = is_decreasing(lagging_df['rolling_slope'].values)

            if is_growingg:
                return 1.
            if is_decreasingg:
                return -1.

            lagging_df['rolling_slope_rank'] = lagging_df['rolling_slope'].rank(pct=True)

            if short:
                lagging_df['rolling_slope_rank_ls'] = 2 * lagging_df['rolling_slope_rank'] - center
                lagging_df['rolling_slope_rank_ls'] = lagging_df['rolling_slope_rank_ls'].clip(-1, 1)
            else:
                lagging_df['rolling_slope_rank_ls'] = lagging_df['rolling_slope_rank']

            gen_sig = lagging_df['rolling_slope_rank_ls'].iloc[-1]
            return gen_sig

        def compute_slope(slope_df):
           y = slope_df.values
           slope = stats.linregress(np.arange(len(y)), y).slope
           return slope

        weekly_df['rolling_slope'] = weekly_df[f'close_{ssj}'].rolling(window=pente_window).apply(compute_slope)


        go = functools.partial(compute_ranked_slope,me_short,me_center)
        signal_df = roll(weekly_df, lookback_window).apply(go)

        signal_df = signal_df.to_frame()
        signal_df.columns = ['signal_gen']

        data_df = pd.merge(data_df.copy(), signal_df.copy(),how='left', right_index=True, left_index=True)
        data_df['signal_gen']=data_df['signal_gen'].ffill()

        data_df['signal'] = data_df['signal_gen'].shift()
        if aggregate_and_save:
            sig_to_save = data_df[[f'close_{ssj}','signal']].copy()
            if aggregated_signals_df is None:
                sig_to_save.columns = [f'close_{ssj}',f'S{ls_signal_suffix}{ssj}']
                aggregated_signals_df = sig_to_save.copy()
            else :
                sig_to_save.columns = [f'close_{ssj}',f'S{ls_signal_suffix}{ssj}']
                aggregated_signals_df = pd.merge(aggregated_signals_df.copy(), sig_to_save.copy(), left_index = True, right_index=True)


        data_df = data_df.dropna()
        data_df['epoch_number'] = data_df['rebalance'].cumsum()

        perf_df = data_df.copy()
        perf_df=perf_df.rename(columns={f'close_{ssj}':'close'})
        freqly_df, _ = signal_utility.reconstitute_signal_perf(data=perf_df, transaction_cost=True,
                                                               normalization=False)

        if plot_subsignal_html:
            fig = realtime_plotting_utility.plot_multiple_time_series(freqly_df[['signal']],logy = False,drop_na_inf = False, title = ssj)
            fig.show()
            fig = realtime_plotting_utility.plot_multiple_time_series(freqly_df[['reconstituted_perf', 'reconstituted_under']],put_on_same_scale = False, drop_na_inf = False, title = ssj)
            fig.show()
            fig1 = realtime_plotting_utility.plot_multiple_time_series(freqly_df[['reconstituted_perf', 'reconstituted_under']],logy = False, put_on_same_scale = False, drop_na_inf = False, title = ssj)
            fig1.show()


        if extract_subsignal_kpi:
            def compute_metrics(df, strat=None):
                kpi_df = riskmetrics.get_kpi(df[[strat]])
                return kpi_df.to_dict()[strat]

            me_strat = 'reconstituted_perf'
            go_comp_kpi = lambda x: compute_metrics(x, strat=me_strat)

            epochkpis_df = freqly_df[['epoch_number', me_strat]].groupby(['epoch_number']).apply(go_comp_kpi)
            epochkpis_df = epochkpis_df.sort_index()
            epochkpis_df = pd.DataFrame.from_records(epochkpis_df.to_dict()).T



            strat_data_df = freqly_df[['reconstituted_perf', 'reconstituted_under']].copy()
            #strat_data_df = strat_data_df[strat_data_df.index >= '2021-02-01'].copy()
            ma = market.MarketAnalyzer(strat_data_df, hourlyze = True)

            print(f'computing kpis for underlying {ssj}')
            kpi_df = ma.get_kpi().T
            kpi_df=kpi_df.dropna()
            kpi_df=kpi_df[~np.isnan(kpi_df.sharpe)]
            kpi_df = kpi_df[kpi_df.index == 'reconstituted_perf']
            print(kpi_df.head())

    def curate_signals(row,under='BTC'):
        if abs(row[f'close_{under}']) <= 1e-3:
            return np.nan
        else:
            return row[f'S{ls_signal_suffix}{ssj}']

    for ssj in ssjs:
        go_curate = lambda x: curate_signals(x, under=ssj)
        aggregated_signals_df[f'S{ls_signal_suffix}{ssj}'] = aggregated_signals_df.apply(go_curate, axis=1)

#    for extra_tok in extra_tokens :
#        aggregated_signals_df[f'S{ls_signal_suffix}{extra_tok}'] = aggregated_signals_df[f'S{ls_signal_suffix}{extra_tok}'].fillna(0.)

    aggregated_signals_df = aggregated_signals_df.dropna()
#    aggregated_signals_df = aggregated_signals_df[~aggregated_signals_df[f'S{ls_signal_suffix}BTC'].isna().values]
#    aggregated_signals_df = aggregated_signals_df.fillna(0.)

    if mix_lo:
        aggregated_signals_df = pd.merge(aggregated_signals_df.copy(), lo_df.copy(),suffixes=['','_lo'], how='left',
                                         left_index=True, right_index=True)
        def compute_mix_signals(row_df, core_tokens= None, extra_tokens= None):
            lo_expo = row_df['expo']
            if lo_expo > 0 :
                return {f'SF{tok}':row_df[f'SMIX{tok}'] for tok in core_tokens + extra_tokens}
            else :
                return {f'SF{tok}':row_df[f'SLS{tok}'] for tok in core_tokens + extra_tokens}

        go_mix = lambda x : compute_mix_signals(x, core_tokens=core_tokens, extra_tokens=extra_tokens)
        nex_signals_series = aggregated_signals_df.apply(go_mix, axis=1)
        nex_signals_df = pd.DataFrame().from_records(nex_signals_series.values)
        nex_signals_df.index = nex_signals_series.index
        aggregated_signals_df = pd.merge(nex_signals_df.copy(),aggregated_signals_df.copy(),left_index=True, right_index=True)
    else :
        for tok in core_tokens + extra_tokens:
            aggregated_signals_df[f'SF{tok}']=aggregated_signals_df[f'SLS{tok}']

    def compute_ls_expo(row_df, universe = None, moderating_factor = np.nan):
        bullish_ones = []
        bearish_ones = []
        bullish_dic = {}
        bearish_dic = {}
        bullish_magnitude = 0.
        bearish_magnitude = 0.
        compo = {}
        for sig in [f'SF{me_tok}' for me_tok in universe]:
            if row_df[sig] > 0. :
                bullish_ones.append(sig)
                bullish_dic[sig] = row_df[sig]
                bullish_magnitude = bullish_magnitude +row_df[sig]
            if row_df[sig]<0.:
                bearish_ones.append(sig)
                bearish_dic[sig] = row_df[sig]
                bearish_magnitude = bearish_magnitude +row_df[sig]

        for sig in [f'SF{me_tok}' for me_tok in universe]:
            if row_df[sig] > 0. and bullish_magnitude>moderating_factor :
                compo[sig] = row_df[sig]/bullish_magnitude*moderating_factor
                continue
            if row_df[sig] > 0.:
                compo[sig] = row_df[sig]
                continue
            if row_df[sig] < 0. and abs(bearish_magnitude)>moderating_factor:
                compo[sig] = row_df[sig]/abs(bearish_magnitude)*moderating_factor
                continue
            if row_df[sig] < 0.:
                compo[sig] = row_df[sig]
                continue
            compo[sig] = 0.

        final_long_expo = 0.
        final_short_expo = 0.
        for sig in [f'SF{me_tok}' for me_tok in universe]:
            if compo[sig]>0.:
                final_long_expo = final_long_expo + compo[sig]
            if compo[sig]<0.:
                final_short_expo = final_short_expo + compo[sig]

        compo['long_expo']  = final_long_expo
        compo['short_expo'] = final_short_expo
        compo['SCASH'] = 1.-compo['long_expo']+ compo['short_expo']
        return compo

    comp_compo = lambda x : compute_ls_expo(x, universe=ssjs, moderating_factor=moderating_factor)
    aggregated_signals_df['compo'] = aggregated_signals_df.apply(comp_compo, axis = 1)

    normed_sig_df = pd.DataFrame().from_records(aggregated_signals_df['compo'].values)
    normed_sig_df.index = aggregated_signals_df.index

    for me_tok in ssjs:
        aggregated_signals_df['return_{}'.format(me_tok)] = aggregated_signals_df['close_{}'.format(me_tok)].pct_change()
    columns_to_keep = [f'return_{me_tok}' for me_tok in ssjs] + [f'close_{me_tok}' for me_tok in ssjs]
    merged_df =  pd.merge(aggregated_signals_df[columns_to_keep].copy(), normed_sig_df.copy(), left_index = True, right_index = True)
    merged_df = merged_df.iloc[1:]
    merged_df.replace([np.inf, -np.inf], np.nan, inplace=True)
    merged_df = merged_df.fillna(0.)

    print('computing exposition')
    def compute_strat_ret(row_df, universe = None, stable_annulalized_apy=0.03):
        aggregated_return = 0.
        for me_token in universe:
            aggregated_return = aggregated_return + row_df[f'return_{me_token}'] * row_df[f'SF{me_token}']
        daily_rate = (1+stable_annulalized_apy)**(1./365.)-1
        aggregated_return = aggregated_return + row_df[f'SCASH']*daily_rate
        return aggregated_return

    comp_ret = lambda x : compute_strat_ret(x, universe=ssjs)
    merged_df['STRAT_RETURN'] = merged_df.apply(comp_ret, axis=1)

    merged_df['COMBO_STRAT'] = np.cumprod(1 + merged_df['STRAT_RETURN'].values)
    merged_df['DOLLAR_COMBO_STRAT'] = 10000. * np.cumprod(1 + merged_df['STRAT_RETURN'].values)
    return merged_df.copy()



import datetime as dt
import pandas as pd
import requests
import json
import math
from dateutil import parser

import numpy as np
from datetime import timedelta
import os

from cryptotoolbox.file_saver import dropbox_file_saver
import time

from cryptotoolbox.bot import binance_bot
from datetime import timedelta

import datetime
# CryptoCompare

sleeping_time = 0.5

def to_unix(s):
    if isinstance(s, str):
        dt_object = parser.parse(s)
    else:
        dt_object = s
    return int(dt_object.replace(tzinfo=dt.timezone.utc).timestamp())


def to_iso(ts):
    if type(ts) == str:
        ts = parser.parse(ts)
        return ts.replace(tzinfo=dt.timezone.utc).isoformat()
    else:
        return dt.datetime.utcfromtimestamp(ts).replace(tzinfo=dt.timezone.utc).isoformat()


def assess_rsp(response):
    if response.status_code != 200:
        raise RuntimeError('Bad gateway:', response.status_code)
    elif isinstance(response.json(), list) and len(response.json()) == 0:
        raise ValueError('No data')
    # elif response.json()['Response'] != 'Success':
    # raise RuntimeError(response.json()['Message'])


def extract_df(optype, precision, r):
    if optype == 'OHLC':
        try:
            tmp = r.json()['Data']['Data']
            return pd.DataFrame(tmp)
        except KeyError:
            print(r.json()['Message'])
    else:
        return pd.DataFrame(r.json()['Data'])


class CryptoCompare:
    def __init__(self, api_key=None, exchange=None):
        self.api_key = api_key
        self.e = exchange

    def get(self, optype, currency, startdate, enddate, precision):
        timedelta = to_unix(enddate) - to_unix(startdate)
        ts = to_unix(enddate)
        precision_dct = {'1h': 3600, 'hour': 3600, 'minute': 60}
        endpoint_dct = {'OHLC': {'url': 'https://min-api.cryptocompare.com/data/v2/histo{}'.format(precision),
                                 'params': {'fsym': currency, 'tsym': 'USD', 'limit': '2000', 'aggreggate': '1',
                                            'toTs': ts}},
                        'OBL2': {'url': 'https://min-api.cryptocompare.com/data/ob/l2/snapshot',
                                 'params': {'api_key', self.api_key}},
                        'HVOL': {'url': 'https://min-api.cryptocompare.com/data/symbol/histohour',
                                 'params': {'fsym': currency, 'tsym': 'USD', 'limit': '500', 'toTs': ts}}}

        if optype == 'OHLC' and precision == 'minute':
            endpoint_dct[optype]['params']['api_key'] = '{' + self.api_key + '}'

        runs, rest = divmod(timedelta / precision_dct[precision], int(endpoint_dct[optype]['params']['limit']))
        runs, rest = int(runs), str(int(math.ceil(rest)))
        output = pd.DataFrame()
        for run in range(runs):
            r = requests.request("GET", endpoint_dct[optype]['url'], params=endpoint_dct[optype]['params'])
            assess_rsp(r)
            output = pd.concat([output, extract_df(optype, precision, r)])
            endpoint_dct[optype]['params'].update({'toTs': output.time.min()})

        if rest != '0':
            endpoint_dct[optype]['params'].update({'limit': rest})
            print(endpoint_dct[optype]['params'])
            r = requests.request("GET", endpoint_dct[optype]['url'], params=endpoint_dct[optype]['params'])
            assess_rsp(r)
            output = pd.concat([output, extract_df(optype, precision, r)])

        output['timestamp'] = output.time.apply(lambda x: to_iso(x))
        output = output.set_index('timestamp', drop=True).sort_index().drop_duplicates()
        return output


def binance_get_histo_future_balance_snapshot(api, p, request_date):
    bot = binance_bot.NaPoleonBinanceFutureBot(api, p)
    return bot.get_futures_balance_snapshot(request_date)

def binance_get_future_balance_history(api, p, start_time, end_time):
    bot = binance_bot.NaPoleonBinanceFutureBot(api, p)
    temp_time = start_time
    valos = None
    while temp_time < end_time:
        temp_time = min(end_time, start_time+timedelta(int(30)))
        print(f'requesting between {start_time} and {temp_time}')
        temp_valos = bot.get_futures_balance_history(start_time, temp_time)
        start_time = temp_time
        if valos is None:
            valos = temp_valos.copy()
        else:
            valos = pd.concat([temp_valos, valos])

    valos['date'] = pd.to_datetime(valos['date'])
    valos = valos.sort_values(by=['date'])
    return valos

    # deltadays = end_time - start_time
    # middletime = start_time+ timedelta(int(deltadays.days/2))
    # first_valos = bot.get_futures_balance_history(start_time, middletime)
    # second_valos = bot.get_futures_balance_history(middletime, end_time)
    # return pd.concat([first_valos,second_valos])


def binance_get_futures_transfer_histo(api, p, start_time, end_time):
    bot = binance_bot.NaPoleonBinanceFutureBot(api, p)
    return bot.get_futures_transfer_history(['USDT', 'BNB'], start_time, end_time)

def binance_get_position_risk(api, p, start_time, end_time):
    bot = binance_bot.NaPoleonBinanceFutureBot(api, p)
    positions_df = pd.DataFrame(bot.get_future_position_risk())
    positions_df = positions_df[abs(positions_df['positionAmt'].astype(float)) > 0].copy()
    return positions_df

def binance_get_position_margin_history(api, p, start_time, end_time):
    bot = binance_bot.NaPoleonBinanceFutureBot(api, p)
    return bot.get_futures_positions_margin_change_history(['BTCUSDT','ETHUSDT'], start_time, end_time)


def request_day_data_paquet(url, me_ts, ssj):
    r = requests.get(url.format(ssj, me_ts))
    dataframe = None
    try:
        dataframe = pd.DataFrame(json.loads(r.text)['Data'])
    except Exception as e:
        print('no data')
    return dataframe

def cryptocompare_get(ssj ='USDT',ssj_against='USD', daily_crypto_starting_day=None, daily_crypto_ending_day=None, aggreg ='1d'):
    assert aggreg == '1d'
    dates_stub = daily_crypto_starting_day.strftime('%d_%b_%Y') + '_' + daily_crypto_ending_day.strftime('%d_%b_%Y')

    year = datetime.datetime.now().year
    month = datetime.datetime.now().month
    day = datetime.datetime.now().day
    hour = datetime.datetime.utcnow().hour
    ts = datetime.datetime(year, month, day, tzinfo=datetime.timezone.utc).timestamp() + hour * 3600
    ts1 = ts - 2001 * 3600 * 24
    ts2 = ts1 - 2001 * 3600 * 24
    ts3 = ts2 - 2001 * 3600 * 24
    ts4 = ts3 - 2001 * 3600 * 24
    ts5 = ts4 - 2001 * 3600 * 24
    ts6 = ts5 - 2001 * 3600 * 24
    ts7 = ts6 - 2001 * 3600 * 24
    ts8 = ts7 - 2001 * 3600 * 24

    print('Loading data')
    day_url_request = 'https://min-api.cryptocompare.com/data/histoday?fsym={}&tsym=' + ssj_against + '&toTs={}&limit=2000'
    dataframe = None
    for me_timestamp in [ts8, ts7, ts6, ts5, ts4, ts3, ts2, ts1, ts]:
        print('waiting')
        df = request_day_data_paquet(day_url_request, me_timestamp, ssj)
        if df is not None:
            if dataframe is not None:
                dataframe = dataframe.append(df, ignore_index=True)
            else:
                dataframe = df.copy()

    dataframe['time'] = pd.to_datetime(dataframe['time'], unit='s')
    dataframe = dataframe.sort_values(by=['time'])
    dataframe = dataframe.rename(columns={"time": "date"}, errors="raise")
    dataframe = dataframe.set_index(dataframe['date'])
    dataframe = dataframe.drop(columns=['date'])
    print('size fetched')
    print(dataframe.shape)
    dataframe = dataframe[dataframe.index >= daily_crypto_starting_day]
    dataframe = dataframe[dataframe.index <= daily_crypto_ending_day]
    print('size filtered after ' + str(daily_crypto_starting_day))
    print(dataframe.shape)
    return dataframe


#https://binance-docs.github.io/apidocs/spot/en/#wallet-endpoints
def binance_get(api, p, type_, start_time, end_time = None, symbol = None, aggreg = None):
  '''
  Accepted types: futures_funding, all_futures_orders, futures_klines (aggreg = 1H), spot_klines (aggreg = '1m', '1h')
  Accepted start_time, end_time: datetime ns
  '''

  from binance.client import Client
  import json
  import time

  client = Client(api, p)
  assert type_ in ['trade_history', 'futures_funding', 'futures_klines', 'wallet_history', 'spot_klines'], 'Wrong type_: trade_history, futures_funding, futures_klines, wallet_history available'
  if aggreg is not None:
    assert aggreg in ['1h', '1m', '5m','30m', '1d','12h'], 'Wrong aggreg_: 1m, 1h available'
    dct_cvt = {'1m': Client.KLINE_INTERVAL_1MINUTE, '5m': Client.KLINE_INTERVAL_5MINUTE, '30m': Client.KLINE_INTERVAL_30MINUTE, '1h':Client.KLINE_INTERVAL_1HOUR, '1d':Client.KLINE_INTERVAL_1DAY, '12h':Client.KLINE_INTERVAL_12HOUR}
    aggreg = dct_cvt[aggreg]

  output = pd.DataFrame()
  if type_ == 'futures_funding' and symbol is not None:
    startTime = int(start_time.timestamp()*1000)
    endTime = int(end_time.timestamp()*1000)
    while startTime<endTime:
      time_col = 'fundingTime'
      try:
        print(f'requesting funding {time_col} between {startTime} and {endTime}')
        print(f'requesting {type(startTime)} {type(endTime)} {startTime} {endTime}')
        startTime =int(startTime)
        endTime=int(endTime)
        output = pd.concat([output, pd.DataFrame(client.futures_funding_rate(symbol = symbol, startTime = startTime, endTime = endTime, limit = 1000))])
        time.sleep(sleeping_time)
      except json.decoder.JSONDecodeError:
        print('Could not convert to Json. Passing @time:', startTime)
      endTime = output.loc[:,time_col].min()
  elif type_ == 'trade_history':
    print('Retrieving all orders in 3H slices. Stay put.')
    time_col = 'time'
    drange = list(pd.date_range(start_time, end_time, freq ='3H'))
    for _ in range(len(drange)-1):
      startTime = int(drange[_].timestamp()*1000)
      endTime = int(drange[_+1].timestamp()*1000)
      try:
        print(f'requesting trades {time_col} between {startTime} and {endTime}')
        time.sleep(sleeping_time)
        print(f'requesting {type(startTime)} {type(endTime)} {startTime} {endTime}')
        startTime =int(startTime)
        endTime=int(endTime)
        tmp = pd.DataFrame(client.futures_account_trades(limit = 1000, startTime = startTime, endTime = endTime))
        output = pd.concat([output, tmp])
        time.sleep(sleeping_time)
      except json.decoder.JSONDecodeError:
        print('Could not convert to Json. Passing @time:', startTime)
        continue
    time.sleep(1)
  elif (type_ == 'futures_klines') and (symbol is not None) and (aggreg is not None):
    startTime = int(start_time.timestamp()*1000)
    endTime = int(end_time.timestamp()*1000)
    previousStartTime = startTime

    while startTime<endTime:
      time_col = 'time'
      try:
        print(f'requesting futures_kline {time_col} between {startTime} and {endTime}')
        time.sleep(sleeping_time)
        print(f'requesting {type(startTime)} {type(endTime)} {startTime} {endTime}')
        startTime =int(startTime)
        endTime=int(endTime)
        output = pd.concat([output, pd.DataFrame(client.futures_klines(symbol = symbol, interval = aggreg, startTime = startTime, endTime = endTime, limit = 1000))])
        time.sleep(sleeping_time)
      except json.decoder.JSONDecodeError:
        print('Could not convert to Json. Passing @time:', startTime)
      #futures_kline
      previousStartTime = startTime
      if output.empty:
          break
      startTime = output.loc[:, 0].max()
      if previousStartTime == startTime :
          break

    output.columns = ['time', 'open', 'high', 'low', 'close', 'volume', 'close_time', 'quote_asset_vol', 'ntrades', 'takerbuy_base_asset_vol', 'takerbuy_quote_asset_vol', 'ignore']
  elif type_ == 'wallet_history':
    startTime = int(start_time.timestamp()*1000)
    endTime = int(end_time.timestamp()*1000)
    previousStartTime = startTime
    while startTime<endTime:
      time_col = 'time'
      try:
        print(f'requesting income history {time_col} between {startTime} and {endTime}')
        time.sleep(sleeping_time)
        print(f'requesting {type(startTime)} {type(endTime)} {startTime} {endTime}')
        startTime =int(startTime)
        endTime=int(endTime)
        output = pd.concat([output, pd.DataFrame(client.futures_income_history(startTime = startTime, endTime=endTime, limit = 1000))])
        time.sleep(sleeping_time)

      except json.decoder.JSONDecodeError:
        print('Could not convert to Json. Passing @time:', startTime)
      # fincome histo
      previousStartTime = startTime
      startTime = output.loc[:, 'time'].max()
      if previousStartTime == startTime :
          break
  elif (type_ == 'spot_klines')  and (symbol is not None) and (aggreg is not None):
    startTime = int(start_time.timestamp()*1000)
    endTime = int(end_time.timestamp()*1000)
    previousStartTime = startTime

    while startTime<endTime:
      time_col = 'time'
      try:
        print(f'requesting income historical kline {time_col} between {startTime} and {endTime}')
        time.sleep(sleeping_time)
        print(f'requesting {type(startTime)} {type(endTime)} {startTime} {endTime}')
        startTime =int(startTime)
        endTime=int(endTime)
        output = pd.concat([output, pd.DataFrame(client.get_historical_klines(symbol = symbol, interval = aggreg, start_str = startTime, end_str = endTime, limit = 500))])
      except json.decoder.JSONDecodeError:
        print('Could not convert to Json. Passing @time:', startTime)
      previousStartTime = startTime
      startTime = output.loc[:, 0].max()
      if previousStartTime == startTime :
          break

    output.columns = ['time', 'open', 'high', 'low', 'close', 'volume', 'close_time', 'quote_asset_vol', 'ntrades', 'takerbuy_base_asset_vol', 'takerbuy_quote_asset_vol', 'ignore']
  else:
    print('Wrong param combination.')
    return
  output['ts'] = (output[time_col] / 1000).apply(lambda x: dt.datetime.utcfromtimestamp(x))
  output = output.sort_values(by='ts', ascending = True).drop_duplicates()
  print(type_, 'done. Shape:', output.shape)
  return output



def fetch_binance_ohlc(symbol = 'BTCUSDT', frequence = '1d', public_key = None, private_key= None, start_time= None, end_time= None, me_month=None, save_to_dropbox = False, dropbox_token = None, local_root_directory = './', trade_only_granularity = False, recompute_all=False, save_to_disk = True):
    print('fetching daily tezos')
    if recompute_all:
        binance_df = binance_get(public_key, private_key, 'spot_klines', start_time, end_time=end_time,
                                           symbol=symbol, aggreg=frequence)

        binance_df.index = pd.to_datetime(binance_df['ts'])
        binance_df['open'] = binance_df['open'].astype(float)
        binance_df['close'] = binance_df['close'].astype(float)
        binance_df['high'] = binance_df['high'].astype(float)
        binance_df['low'] = binance_df['low'].astype(float)
        if save_to_disk:
            save_equity_daily_data(data_df = binance_df, ticker = symbol, frequence=frequence, starting_date=start_time, running_date = end_time, local_root_directory = local_root_directory)
    else :
        try :
            filename = symbol + '_' + frequence + '_' + start_time.strftime('%d_%b_%Y') + '_' + end_time.strftime(
                '%d_%b_%Y') + '_daily_returns.pkl'
            binance_df = pd.read_pickle(local_root_directory + filename)
        except Exception as e :
            print(f'cant fetch data {e}')
            binance_df = binance_get(public_key, private_key, 'spot_klines', start_time, end_time=end_time,
                                     symbol=symbol, aggreg=frequence)

            binance_df.index = pd.to_datetime(binance_df['ts'])
            binance_df['open'] = binance_df['open'].astype(float)
            binance_df['close'] = binance_df['close'].astype(float)
            binance_df['high'] = binance_df['high'].astype(float)
            binance_df['low'] = binance_df['low'].astype(float)
            if save_to_disk:
                save_equity_daily_data(data_df=binance_df, ticker=symbol, frequence=frequence, starting_date=start_time,
                                   running_date=end_time, local_root_directory=local_root_directory)

    return binance_df

def save_equity_daily_data(data_df = None, ticker=None, frequence=None, provider = 'napoleon', starting_date=None,running_date=None,local_root_directory=None):
    filename = ticker + '_' + frequence + '_' + starting_date.strftime('%d_%b_%Y') + '_' + running_date.strftime(
        '%d_%b_%Y') + '_daily_returns.pkl'
    data_df.to_pickle(local_root_directory + filename)







from datetime import datetime

from cryptotoolbox.connector import crypto_connector
from cryptotoolbox.realtime import realtime_plotting_utility
from scipy.interpolate import interp1d

from gql import gql, Client
from gql.transport.requests import RequestsHTTPTransport
import numpy as np

import pandas as pd


import datetime as datetime


from cryptotoolbox.connector import crypto_connector
from scipy import stats
import pandas as pd
import numpy as np
from numpy.lib.stride_tricks import as_strided as stride
from cryptotoolbox.realtime import realtime_plotting_utility
from scipy.optimize import Bounds, LinearConstraint, minimize



def roll(df, w):
    v = df.values
    d0, d1 = v.shape
    s0, s1 = v.strides
    restricted_length = d0 - (w - 1)
    a = stride(v, (restricted_length, w, d1), (s0, s0, s1))
    rolled_df = pd.concat({
        row: pd.DataFrame(values, columns=df.columns)
        for row, values in zip(df.index[-restricted_length:], a)
    })
    return rolled_df.groupby(level=0)



def annualize_percent(first_value=np.nan, last_value=np.nan, nb_years=np.nan):
    return (last_value / first_value) ** (1. / nb_years) - 1.


def compute_impermanent_loss_v2(p_0=np.nan, p_out=np.nan):
    tau = p_out / p_0
    return 2 * np.sqrt(tau) / (1. + tau) - 1, tau


def compute_impermanent_loss_v3(p_0=np.nan, p_a=np.nan, p_b=np.nan, p_out=np.nan):
    ILv2, tau = compute_impermanent_loss_v2(p_0=p_0, p_out=p_out)
    square_one = np.sqrt(p_a / p_0)
    square_two = np.sqrt(p_0 / p_b)
    num_one = square_one + tau * square_two
    factor = 1. / (1. - num_one / (1. + tau))
    ILv3 = ILv2 * factor
    return ILv3, ILv2

### formule venant d'ici (à redémontrer)
#https://lambert-guillaume.medium.com/understanding-the-value-of-uniswap-v3-liquidity-positions-cdaaee127fe7
### formule venant d'ici (à redémontrer)
#https://lambert-guillaume.medium.com/pricing-uniswap-v3-lp-positions-towards-a-new-options-paradigm-dce3e3b50125
#https://lambert-guillaume.medium.com/how-to-create-a-perpetual-options-in-uniswap-v3-3c40007ccf1

def compute_total_LP_portfolio_value_first_methodo(p_a=np.nan, p_b=np.nan, p_out=np.nan):
    first_size = np.sqrt(p_a * p_b) * (np.sqrt(p_out) - np.sqrt(p_a))/(np.sqrt(p_b) - np.sqrt(p_a))
    second_size = np.sqrt(p_a * p_out) * (np.sqrt(p_b) - np.sqrt(p_out))/(np.sqrt(p_b) - np.sqrt(p_a))
    return first_size + second_size

def compute_total_LP_portfolio_value_second_methodo(p_a=np.nan, p_b=np.nan, p_out=np.nan):
    K = np.sqrt(p_a * p_b)
    r = np.sqrt(p_b/p_a)
    return (2.*np.sqrt(p_out * K * r) - p_out - K)/(r-1)

def getBBands(df, period=10, stdNbr=2):
    df['middle'] = get_sma(df['Close'], period)
    df['std'] = df['Close'].rolling(period).std()
    df['upper'] = df['middle'] + df['std'] * stdNbr
    df['lower'] = df['middle'] - df['std'] * stdNbr
    return df


def get_sma(prices, rate):
    return prices.rolling(rate).mean()


def get_bollinger_bands(prices, rate=20):
    sma = get_sma(prices, rate)
    std = prices.rolling(rate).std()
    bollinger_up = sma + std * 2  # Calculate top band
    bollinger_down = sma - std * 2  # Calculate bottom band
    return bollinger_up, bollinger_down


def getBBands(df, period=10, stdNbr=2):
    df['middle'] = get_sma(df['Close'], period)
    df['std'] = df['Close'].rolling(period).std()
    df['upper'] = df['middle'] + df['std'] * stdNbr
    df['lower'] = df['middle'] - df['std'] * stdNbr
    return df


def get_sma(prices, rate):
    return prices.rolling(rate).mean()


def get_bollinger_bands(prices, rate=20):
    sma = get_sma(prices, rate)
    std = prices.rolling(rate).std()
    bollinger_up = sma + std * 2  # Calculate top band
    bollinger_down = sma - std * 2  # Calculate bottom band
    return bollinger_up, bollinger_down

def get_price_data(me_pair=None, initial_date=None, local_root_directory='none'):
    ssj = me_pair[0]
    ssj_against = me_pair[1]

    pair = ssj + ssj_against
    daily_crypto_starting_day = initial_date

    daily_crypto_starting_date = datetime.datetime.strptime(daily_crypto_starting_day, '%Y-%m-%d')
    starting_date = daily_crypto_starting_date
    running_date = datetime.datetime.now()

    nb_days_total = (running_date - daily_crypto_starting_date).days
    print(nb_days_total)
    refetch_all = True
    data_df = crypto_connector.fetch_crypto_daily_data(ssj=ssj, ssj_against=ssj_against,
                                                       local_root_directory=local_root_directory,
                                                       daily_return_pkl_filename_suffix='_daily_returns.pkl',
                                                       refetch_all=refetch_all,
                                                       daily_crypto_starting_day=daily_crypto_starting_date,
                                                       daily_crypto_ending_day=running_date)

    data_df = data_df[['open', 'high', 'low', 'close', 'volume']]
    data_df.columns = ['Open', 'High', 'Low', 'Close', 'volume']
    data_df['Date'] = data_df.index
    data_df['DateStr'] = data_df.index.astype(str)
    print('done')
    data_df['Date'] = pd.to_datetime(data_df['Date'])
    data_df['weekday'] = data_df.apply(lambda x: datetime.datetime.date(x['Date']).weekday(), axis=1)
    data_df = data_df.set_index('Date')

    return data_df

def get_apr(date, pct_change, volume_df = None, current_apr=None):
    x = np.array(list(current_apr.keys()))
    y = list(current_apr.values())
    f2 = interp1d(x, y, kind='linear')
    try:
        pct_change = max(1.,pct_change)
        apr = f2(pct_change)
    except Exception as e:
        print(f'trouble {pct_change}')
        print(e)
    current_apr_ethbtc = apr/100.
    daily_current_apr_ethbtc= (1.+current_apr_ethbtc)**(1./365.)-1.

    ratio = volume_df.loc[date].volumeToken0 / volume_df.iloc[-1].volumeToken0
    prorata_daily_current_apr_ethbtc = daily_current_apr_ethbtc * ratio
    return prorata_daily_current_apr_ethbtc

def get_swapping_cost(swapping_size_proportion =None, swapping_cost_map=None):
    if np.isnan(swapping_size_proportion):
        return np.nan
    assert swapping_size_proportion <= 1.
    assert swapping_size_proportion >= 0.
    percentage_swapping_size_proportion = 100. * swapping_size_proportion
    x = np.array(list(swapping_cost_map.keys()))
    y = list(swapping_cost_map.values())
    f2 = interp1d(x, y, kind='linear')
    cost = np.nan
    try:
        cost = f2(percentage_swapping_size_proportion).item()
    except Exception as e:
        print(f'trouble computing swapping cost{percentage_swapping_size_proportion}')
        print(e)
    return cost



def LS_IL_boundaries_with_signal(data_df = None, plotHtml = False, me_pair = ('ETH', 'BTC'), bb_periods=[5, 15, 30, 55], stdNbrs=[2., 3., 5.], distance_to_bound_percs = [0,5,10,15,20,30,40], current_apr=None, bypass_signal=True, bypass_stoploss=True, swapping_cost_map = {1:1e-4, 10:0.01, 50:0.05, 75:0.075, 100:0.1},starting_number_of_numeraire = 1000):
    ssj = me_pair[0]
    ssj_against = me_pair[1]
    objectives = []

    for bb_period in bb_periods:
        for stdNbr in stdNbrs:
            for distance_to_bound_perc in distance_to_bound_percs:
                df = data_df.copy()
                df['Date']=df.index
                print(f'params_{bb_period}_{stdNbr}')
                df = getBBands(df.copy(), period=bb_period, stdNbr=stdNbr)
                df = df.dropna()

                stage_nb = 0

                pct_change = np.nan
                real_upper_bound = np.nan
                real_lower_bound = np.nan
                buffer_upper_bound = np.nan
                buffer_lower_bound = np.nan

                initial_real_upper_bound = np.nan
                initial_real_lower_bound = np.nan


                df['pct_change'] = (stdNbr * df['std']) / df['Close'] * 100
                avg_pct_change = df['pct_change'].mean()
                ### to compute a proper metric
                if bypass_stoploss:
                    df['stoploss_condition'] = False
                if bypass_signal:
                    df['signal'] = 1.

                if plotHtml:
                    fig = realtime_plotting_utility.plot_multiple_time_series(
                        data_df=df[['Close', 'upper', 'lower']].copy(), logy=False, split=False,
                        put_on_same_scale=False,
                        title=f'Bollinger {ssj}{ssj_against}')
                    fig.show()

                initial_price = df['Close'].iloc[0]
                begining_initial_price = df['Close'].iloc[0]

                initial_lp_value = np.nan
                preceding_vault_signal = df['signal'].iloc[0]

                stages = []
                counter = 0
                the_end = False
                for i, row in df.iterrows():
                    if i == df.index[-1]:
                        the_end = True
                    if counter == 0:
                        counter = counter + 1
                        continue
                    ILv3, ILv2 = np.nan, np.nan
                    lp_value_first, lp_value_second = np.nan, np.nan
                    lp_position_profit = np.nan
                    flipped_lp_position_profit = np.nan


                    apr = get_apr(row['DateStr'],  row['pct_change'], volume_df=data_df.copy(), current_apr=current_apr)

                    # if currently providing and we enter in a risk off mode we need to compute IL
                    we_are_out = np.isnan(real_upper_bound)

                    stoploss_condition = row['stoploss_condition']
                    vault_signal = row['signal']

                    print(f'real_upper_bound {real_upper_bound}')
                    high = row['High']
                    print(f'high {high}')
                    print(f'real_lower_bound {real_lower_bound}')
                    low = row['Low']
                    print(f'low {low}')
                    print(f'stoploss_condition {stoploss_condition}')
                    print(f'distance to bounds {distance_to_bound_perc}')
                    curILv3 = np.nan
                    curILv2 = np.nan
                    curlp_value_first = np.nan
                    curlp_value_second = np.nan
                    curlp_position_profit = np.nan
                    curflipped_lp_position_profit = np.nan

                    # defining_new_state = (bounds_broken and providing) or (providing and slope_too_sloppy)
                    broken_bounds = np.nan
                    if not we_are_out:
                        distance_to_bound = distance_to_bound_perc/100. * real_upper_bound
                        buffer_upper_bound = real_upper_bound - distance_to_bound
                        buffer_lower_bound = real_lower_bound + distance_to_bound

                        broken_bounds= (row['High'] > buffer_upper_bound or row['Low'] < buffer_lower_bound)
                        assert (not np.isnan(initial_price)) and (not np.isnan(real_lower_bound)) and (
                            not np.isnan(real_upper_bound))
                        p_out = row['Close']
                        curILv3, curILv2 = compute_impermanent_loss_v3(p_0=initial_price, p_a=real_lower_bound,
                                                                 p_b=real_upper_bound,
                                                                 p_out=p_out)
                        curlp_value_first = compute_total_LP_portfolio_value_first_methodo(p_a=real_lower_bound,
                                                                 p_b=real_upper_bound,
                                                                 p_out=p_out)
                        curlp_value_second = compute_total_LP_portfolio_value_second_methodo(p_a=real_lower_bound,
                                                                 p_b=real_upper_bound,
                                                                 p_out=p_out)
                        curlp_position_profit = (curlp_value_first - initial_lp_value) / initial_lp_value
                        curflipped_lp_position_profit = preceding_vault_signal * curlp_position_profit



                    signal_change = False
                    if vault_signal != preceding_vault_signal:
                        signal_change = True

                    we_enter = False
                    we_out = False
                    we_rebalance_signal = False
                    we_rebalance_bounds = False
                    #############
                    if stoploss_condition:
                        if we_are_out: ### we do nothing
                            print('we do nothing, already out')
                        if not we_are_out: ### we out
                            print('we out the position because of the stop loss condition')
                            we_out = True
                            assert (not np.isnan(initial_price)) and (not np.isnan(real_lower_bound)) and (
                                not np.isnan(real_upper_bound))
                            p_out = row['Close']
                            ILv3, ILv2 = compute_impermanent_loss_v3(p_0=initial_price, p_a=real_lower_bound,
                                                                     p_b=real_upper_bound,
                                                                     p_out=p_out)
                            lp_value_first = compute_total_LP_portfolio_value_first_methodo(p_a=real_lower_bound,
                                                                                            p_b=real_upper_bound,
                                                                                            p_out=p_out)
                            lp_value_second = compute_total_LP_portfolio_value_second_methodo(p_a=real_lower_bound,
                                                                                             p_b=real_upper_bound,
                                                                                             p_out=p_out)
                            lp_position_profit = (lp_value_first-initial_lp_value)/initial_lp_value
                            flipped_lp_position_profit = preceding_vault_signal * lp_position_profit

                            real_upper_bound = np.nan
                            real_lower_bound = np.nan
                            stage_nb = stage_nb + 1

                    if not stoploss_condition:
                        if we_are_out: ### we enter a position  cause no stop loss signal
                            we_enter = True
                            print('we enter')
                            initial_price = row['Close']
                            real_upper_bound = row['High'] + stdNbr * row['std']

                            pct_change = (real_upper_bound - row['High']) / row['High']
                            real_lower_bound = row['Low'] - stdNbr * row['std']

                            if np.isnan(initial_real_upper_bound):
                                initial_real_upper_bound = real_upper_bound
                                initial_real_lower_bound = real_lower_bound

                            stage_nb = stage_nb + 1
                            if np.isnan(initial_lp_value):
                                initial_lp_value = compute_total_LP_portfolio_value_first_methodo(p_a=real_lower_bound,
                                                                                                  p_b=real_upper_bound,
                                                                                                  p_out=initial_price)

                        if not we_are_out:
                            if broken_bounds or signal_change or the_end:## we rebalance
                                if broken_bounds:
                                    we_rebalance_bounds = True
                                    print('we rebalance because of buffer bounds being touched')
                                if signal_change:
                                    we_rebalance_signal=True
                                    print('we rebalance because of long/short signal change')
                                up = row['High'] > buffer_upper_bound
                                down = row['Low'] < buffer_lower_bound
                                ## one bound must be broken
                                assert up or down or signal_change or the_end
                                # p_out = np.nan
                                # if up:
                                #     p_out = real_upper_bound
                                # elif down:
                                #     p_out = real_lower_bound
                                # else:
                                p_out = row['Close']

                                ##### all in LP loss
                                ILv3, ILv2 = compute_impermanent_loss_v3(p_0=initial_price, p_a=real_lower_bound,
                                                                         p_b=real_upper_bound,
                                                                         p_out=p_out)
                                lp_value_first = compute_total_LP_portfolio_value_first_methodo(p_a=real_lower_bound,
                                                                                                p_b=real_upper_bound,
                                                                                                p_out=p_out)
                                lp_value_second = compute_total_LP_portfolio_value_second_methodo(p_a=real_lower_bound,
                                                                                                  p_b=real_upper_bound,
                                                                                                  p_out=p_out)

                                lp_position_profit = (lp_value_first - initial_lp_value) / initial_lp_value
                                flipped_lp_position_profit = preceding_vault_signal * lp_position_profit


                                # if up:
                                #     initial_price = row['High']
                                # else:
                                #     initial_price = row['Low']
                                initial_price = row['Close']

                                # here compute the previous impermanent loss

                                stage_nb = stage_nb + 1
                                real_upper_bound = row['High'] + stdNbr * row['std']

                                pct_change = (real_upper_bound - row['High']) / row['High']
                                real_lower_bound = row['Low'] - stdNbr * row['std']

                                initial_lp_value = compute_total_LP_portfolio_value_first_methodo(p_a=real_lower_bound,
                                                                                                  p_b=real_upper_bound,
                                                                                                  p_out=p_out)

                    stages.append({
                        'date': row['Date'],
                        'Close':row['Close'],
                        'incurred_v3': ILv3,
                        'incurred_v2': ILv2,
                        'pct_change' : pct_change,
                        'lp_position_profit':lp_position_profit,
                        'flipped_lp_position_profit':flipped_lp_position_profit,
                        'lp_value_first': lp_value_first,
                        'lp_value_second': lp_value_second,
                        'initial_lp_value':initial_lp_value,
                        'initial_price': initial_price,
                        'signal_change':signal_change,
                        'preceding_vault_signal':preceding_vault_signal,
                        'broken_bounds':broken_bounds,
                        'stoploss_condition':stoploss_condition,
                        'the_end': the_end,
                        'we_enter': we_enter,
                        'we_out' : we_out,
                        'curILv3' :curILv3,
                        'curILv2' :curILv2,
                        'curflipped_lp_position_profit': curflipped_lp_position_profit,
                        'curlp_position_profit': curlp_position_profit,
                        'curlp_value_first' : curlp_value_first,
                        'curlp_value_second' : curlp_value_second,
                        'we_rebalance_signal' : we_rebalance_signal,
                        'we_rebalance_bounds' : we_rebalance_bounds,
                        'real_upper_bound': real_upper_bound,
                        'real_lower_bound': real_lower_bound,
                        'buffer_upper_bound': buffer_upper_bound,
                        'buffer_lower_bound': buffer_lower_bound,
                        'distance_to_bound_perc':distance_to_bound_perc,
                        'stage_nb': stage_nb,
                        'apr': apr,
                    })
                    preceding_vault_signal = vault_signal

                stages_df = pd.DataFrame(stages)

                ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### #######
                ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### #######
                ####### ####### ####### ####### swapping cost analysis
                ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### #######
                ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### #######
                def rebalancing_signal(row):
                    we_rebalance_signal = row['we_rebalance_signal']
                    we_rebalance_bound = row['we_rebalance_bounds']
                    return we_rebalance_bound or we_rebalance_signal

                stages_df['rebalancing'] = stages_df.apply(rebalancing_signal, axis=1)

                rebalancing_df = stages_df[stages_df['rebalancing']].copy()

                if len(rebalancing_df)>0:
                    rebalancing_df['leaving_initial_price'] = rebalancing_df['initial_price'].shift(+1)
                    rebalancing_df['leaving_initial_price'].iloc[0] = begining_initial_price
                    rebalancing_df['leaving_real_upper_bound'] = rebalancing_df['real_upper_bound'].shift(+1)
                    rebalancing_df['leaving_real_lower_bound'] = rebalancing_df['real_lower_bound'].shift(+1)
                    rebalancing_df['leaving_real_upper_bound'].iloc[0] = initial_real_upper_bound
                    rebalancing_df['leaving_real_lower_bound'].iloc[0] = initial_real_lower_bound

                    rebal_data_df = rebalancing_df[['date','leaving_initial_price', 'initial_price', 'real_upper_bound', 'real_lower_bound', 'leaving_real_upper_bound', 'leaving_real_lower_bound']]

                    def compute_percentage_spread(row):
                        pct_chagne = (row['leaving_real_upper_bound']-row['leaving_initial_price'])/row['leaving_initial_price']
                        return pct_chagne

                    rebal_data_df['pct_change_upper'] = rebal_data_df.apply(compute_percentage_spread, axis=1)

                    ### ether = X (risqué)
                    ### BTC = Y (numéraire)
                    ### P = P ether in BTC
                    def get_liquidity_from_BTC(nb_amount_BTC, current_price, lower_bound ):
                        return nb_amount_BTC/(np.sqrt(current_price) - np.sqrt(lower_bound) )

                    def get_liquidity_from_ETH(nb_amount_ETH, current_price, upper_bound ):
                        return nb_amount_ETH/(1./np.sqrt(current_price) - 1./np.sqrt(upper_bound))

                    def get_ether_from_liquidity(liquidity, current_price, upper_bound):
                        return liquidity*(1./np.sqrt(current_price) -1./np.sqrt(upper_bound))

                    def get_btc_from_liquidity(liquidity, current_price, lower_bound):
                        return liquidity*(np.sqrt(current_price) - np.sqrt(lower_bound) )




                    previous_liquidities = np.zeros(len(rebal_data_df))
                    previous_ethers = np.zeros(len(rebal_data_df))
                    previous_btcs = np.zeros(len(rebal_data_df))
                    new_liquidities = np.zeros(len(rebal_data_df))
                    new_ethers = np.zeros(len(rebal_data_df))
                    new_btcs = np.zeros(len(rebal_data_df))
                    swapping_sizes = np.zeros(len(rebal_data_df))
                    swap_costs = np.zeros(len(rebal_data_df))
                    count = 0

                    previous_btc = np.nan
                    previous_liquidity = np.nan
                    previous_eth = np.nan

                    new_btc = np.nan
                    new_liquidity = np.nan
                    new_eth = np.nan

                    rebal_data_df.index = range(len(rebal_data_df))
                    to_keep_init_liq = np.nan
                    to_keep_init_eth = np.nan
                    to_keep_init_btc = np.nan
                    to_keep_init_price = np.nan
                    for ii, row in rebal_data_df.iterrows():
                        if ii == 0 :
                            ### btc at initial date
                            previous_btc = starting_number_of_numeraire
                            ## initial position configuration: leaving_initial_price inbetween leaving_real_upper_bound and leaving_real_lower_bound
                            leaving_initial_price = row['leaving_initial_price']
                            leaving_real_upper_bound = row['leaving_real_upper_bound']
                            leaving_real_lower_bound = row['leaving_real_lower_bound']
                            ## liquidity at initial date
                            previous_liquidity = get_liquidity_from_BTC(previous_btc,leaving_initial_price,leaving_real_lower_bound )
                            ## ether at initial date
                            previous_eth = get_ether_from_liquidity(previous_liquidity, leaving_initial_price,leaving_real_upper_bound)
                            previous_liquidity_bis = get_liquidity_from_ETH(previous_eth, leaving_initial_price,leaving_real_upper_bound)
                            to_keep_init_liq=previous_liquidity
                            to_keep_init_eth=previous_eth
                            to_keep_init_btc = previous_btc
                            to_keep_init_price = leaving_initial_price

                            assert abs(previous_liquidity_bis - previous_liquidity) < 1e-3
                        else :
                            previous_btc = new_btc
                            previous_eth = new_eth
                            previous_liquidity = new_liquidity

                        swapping_size = np.nan
                        # we leave the 'leaving position' at a new price which will become the new initial price
                        newly_initial_price = row['initial_price']
                        ### we limit the price move to the upper and lower bounds of the position we are in (quantities do not move after in univ3)
                        truncated_newly_initial_price = min(newly_initial_price,leaving_real_upper_bound )
                        truncated_newly_initial_price = max(truncated_newly_initial_price,leaving_real_lower_bound )

                        def get_new_btc_quantity_from_price_change(liquidity, new_price, old_price):
                            return liquidity * (np.sqrt(new_price)-np.sqrt(old_price))

                        def get_new_ether_quantity_from_price_change(liquidity, new_price, old_price):
                            return liquidity * (1./np.sqrt(new_price)-1./np.sqrt(old_price))

                        deltaBTC = get_new_btc_quantity_from_price_change(previous_liquidity, truncated_newly_initial_price, leaving_initial_price)
                        deltaETH = get_new_ether_quantity_from_price_change(previous_liquidity, truncated_newly_initial_price, leaving_initial_price)

                        previous_ethers[count] = previous_eth
                        previous_btcs[count] = previous_btc
                        previous_liquidities[count] = previous_liquidity

                        after_withdrawal_btc = previous_btc + deltaBTC
                        after_withdrawal_eth = previous_eth + deltaETH

                        print(f'after_withdrawal_eth {after_withdrawal_eth}')
                        print(f'after_withdrawal_btc {after_withdrawal_btc}')

                        ####### rebalancing quantities for the new position
                        new_real_lower_bound = row['real_lower_bound']
                        new_real_upper_bound = row['real_upper_bound']
                        # rebalancing the bounds

                        def get_btc_matching_proportion(eth_amount=None, current_price=None, lower_bound=None,
                                                        upper_bound=None):
                            new_liquidity = get_liquidity_from_ETH(eth_amount, current_price, upper_bound)
                            new_btc = get_btc_from_liquidity(new_liquidity, current_price, lower_bound)
                            return new_btc, new_liquidity

                        def get_eth_matching_proportion(btc_amount=None, current_price=None, lower_bound=None,
                                                        upper_bound=None):
                            new_liquidity = get_liquidity_from_BTC(btc_amount, current_price, lower_bound)
                            new_eth = get_ether_from_liquidity(new_liquidity, current_price, upper_bound)
                            return new_eth, new_liquidity

                        new_btc, _ = get_btc_matching_proportion(eth_amount=after_withdrawal_eth,
                                                                 current_price=newly_initial_price,
                                                                 lower_bound=new_real_lower_bound,
                                                                 upper_bound=new_real_upper_bound)
                        ##### we end up with eth equivalent to more btc than we have : we must swap eth
                        has_to_swap_eth =  new_btc >= after_withdrawal_btc

                        new_eth, _ = get_eth_matching_proportion(btc_amount=after_withdrawal_btc,
                                                                 current_price=newly_initial_price,
                                                                 lower_bound=new_real_lower_bound,
                                                                 upper_bound=new_real_upper_bound)
                        ##### we end up with btc equivalent to more eth than we have : we must swap btc
                        has_to_swap_btc = new_eth >= after_withdrawal_eth
                        print(has_to_swap_btc)
                        print(has_to_swap_eth)
                        if after_withdrawal_btc>0. and after_withdrawal_eth>0.:
                            assert has_to_swap_btc == ~has_to_swap_eth

                        if has_to_swap_btc: #### we have more btcs than eth
                            new_btc = after_withdrawal_btc
                            print(f'new btc {new_btc}')
                            print(f'new price {newly_initial_price}')
                            print(f'new lower bound {new_real_lower_bound}')
                            print(f'new upper bound {new_real_upper_bound}')


                            pool_ratio_price = new_btc/new_eth
                            print(f'new ratio {pool_ratio_price}')

                            def get_eth_amount_to_swap(x= None, slippage = 20e-4, initial_btc=None,initial_ether=None,lower_bound=None, upper_bound=None, current_price=None):
                                btc_left = (1.-x)*initial_btc
                                #### if we swap too big quantities we incur price slippage
                                #### if we swap BTCs to ETH, the price of ETH in BTC Y/X will drop
                                slippage_price = current_price * (1. - slippage)
                                ##### the left btc and swapped ether will make the position
                                eth_in_lp, liquidity = get_eth_matching_proportion(btc_amount=btc_left,
                                                                           current_price=slippage_price,
                                                                           lower_bound=lower_bound,
                                                                           upper_bound=upper_bound)
                                return eth_in_lp ,slippage_price,liquidity

                            def get_eth_amount_to_swap_functional(x= None, slippage = 20e-4, initial_btc=None,initial_ether=None,lower_bound=None, upper_bound=None, current_price=None):
                                btc_left = (1.-x)*initial_btc

                                btc_to_swap = x*initial_btc
                                #### if we swap too big quantities we incur price slippage
                                #### if we swap BTCs to ETH, the price of ETH in BTC Y/X will drop
                                slippage_price = current_price * (1. - slippage)
                                ##### the left btc and swapped ether will make the position
                                eth_in_lp, _ = get_eth_matching_proportion(btc_amount=btc_left,
                                                                           current_price=slippage_price,
                                                                           lower_bound=lower_bound,
                                                                           upper_bound=upper_bound)
                                differential = (eth_in_lp -initial_ether)*slippage_price - btc_to_swap
                                return abs(differential)

                            to_optimize = lambda x: get_eth_amount_to_swap_functional(x=x,initial_btc = after_withdrawal_btc, initial_ether = after_withdrawal_eth,current_price= newly_initial_price, lower_bound=new_real_lower_bound, upper_bound = new_real_upper_bound)
                            test_opti = to_optimize(0.5)
                            low_bound = 0.
                            up_bound = 1.
                            const_ind = Bounds(low_bound , up_bound )


                            result_final = minimize(
                                to_optimize,
                                0.5,
                                bounds=const_ind,
                                method='Nelder-Mead'
                            )
                            optimal_size_to_swap = np.nan
                            if result_final.success:
                                optimal_size_to_swap = result_final.x[0]

                            print(f'optimal_size_to_swap {optimal_size_to_swap}')
                            swapping_size = optimal_size_to_swap
                            new_btc = (1.-optimal_size_to_swap)*after_withdrawal_btc
                            new_eth,slippage_price,liquidity = get_eth_amount_to_swap(x=optimal_size_to_swap,initial_btc = after_withdrawal_btc,  initial_ether = after_withdrawal_eth,current_price= newly_initial_price, lower_bound=new_real_lower_bound, upper_bound = new_real_upper_bound)

                            new_liquidity = get_liquidity_from_ETH(new_eth, slippage_price,new_real_upper_bound)
                            new_liquidity_bis = get_liquidity_from_BTC(new_btc, slippage_price,new_real_lower_bound)

                            if not np.isnan(new_liquidity) and not abs(new_liquidity) < 1e-6:
                                condition_to_check = abs(new_liquidity-new_liquidity_bis)/abs(new_liquidity)<1e-3
                                if not condition_to_check:
                                    print('trouble')
                                assert abs(new_liquidity-new_liquidity_bis)/abs(new_liquidity)<1e-3
                            else :
                                print('trouble')

                        elif has_to_swap_eth :
                            new_eth = after_withdrawal_eth
                            print(f'new eth {new_eth}')
                            print(f'new price {newly_initial_price}')
                            print(f'new lower bound {new_real_lower_bound}')
                            print(f'new upper bound {new_real_upper_bound}')


                            def get_btc_amount_to_swap(x= None, slippage = 20e-4, initial_eth=None, initial_btc=None,lower_bound=None, upper_bound=None, current_price=None):
                                eth_left = (1.-x)*initial_eth
                                eth_to_swap = x*initial_eth
                                #### if we swap too big quantities we incur price slippage
                                #### if we swap BTCs to ETH, the price of ETH in BTC Y/X will drop
                                slippage_price = current_price * (1. + slippage)
                                ##### the left btc and swapped ether will make the position
                                btc_in_lp, liquidity = get_btc_matching_proportion(eth_amount=eth_left,
                                                                           current_price=slippage_price,
                                                                           lower_bound=lower_bound,
                                                                           upper_bound=upper_bound)
                                return btc_in_lp, slippage_price, liquidity

                            def get_btc_amount_to_swap_functional(x= None, slippage = 20e-4, initial_eth=None,initial_btc=None,lower_bound=None, upper_bound=None, current_price=None):
                                eth_left = (1.-x)*initial_eth
                                eth_to_swap = x*initial_eth
                                #### if we swap too big quantities we incur price slippage
                                #### if we swap BTCs to ETH, the price of ETH in BTC Y/X will drop
                                slippage_price = current_price * (1. + slippage)
                                ##### the left btc and swapped ether will make the position
                                btc_in_lp, _ = get_btc_matching_proportion(eth_amount=eth_left,
                                                                           current_price=slippage_price,
                                                                           lower_bound=lower_bound,
                                                                           upper_bound=upper_bound)
                                differential = eth_to_swap*slippage_price - (btc_in_lp-initial_btc)
                                return abs(differential)

                            to_optimize = lambda x: get_btc_amount_to_swap_functional(x=x,initial_eth = after_withdrawal_eth,initial_btc= after_withdrawal_btc, current_price=newly_initial_price, lower_bound=new_real_lower_bound, upper_bound = new_real_upper_bound)
                            test_opti = to_optimize(0.5)
                            low_bound = 0.
                            up_bound = 1.
                            const_ind = Bounds(low_bound , up_bound )


                            result_final = minimize(
                                to_optimize,
                                0.5,
                                bounds=const_ind,
                                method='Nelder-Mead'
                            )
                            optimal_size_to_swap = np.nan
                            if result_final.success:
                                optimal_size_to_swap = result_final.x[0]

                            print(f'optimal_size_to_swap {optimal_size_to_swap}')
                            swapping_size = optimal_size_to_swap
                            new_eth = (1.-optimal_size_to_swap)*after_withdrawal_eth
                            new_btc, slippage_price, liquidity = get_btc_amount_to_swap(x=optimal_size_to_swap,initial_eth = after_withdrawal_eth,initial_btc= after_withdrawal_btc, current_price=newly_initial_price, lower_bound=new_real_lower_bound, upper_bound = new_real_upper_bound)

                            new_liquidity = get_liquidity_from_ETH(new_eth, slippage_price,new_real_upper_bound)
                            new_liquidity_bis = get_liquidity_from_BTC(new_btc, slippage_price,new_real_lower_bound)
                            if not np.isnan(new_liquidity) and not abs(new_liquidity) < 1e-6:
                                condition_to_check = abs(new_liquidity-new_liquidity_bis)/abs(new_liquidity)<1e-3
                                if not condition_to_check:
                                    print('trouble')
                                assert abs(new_liquidity-new_liquidity_bis)/abs(new_liquidity)<1e-3
                            else :
                                print('trouble')
                        else :
                            new_eth = after_withdrawal_eth
                            new_btc = after_withdrawal_btc

                        new_ethers[count] = new_eth
                        new_btcs[count] = new_btc
                        new_liquidities[count] = new_liquidity
                        swapping_sizes[count] = swapping_size
                        swap_cost = get_swapping_cost(swapping_size_proportion=swapping_size, swapping_cost_map=swapping_cost_map)
                        swap_costs[count] = swap_cost

                        previous_liquidity=new_liquidity
                        previous_btc=new_btc
                        previous_eth=new_eth

                        count = count + 1
                        print('done')

                    rebal_data_df['previous_btcs'] = previous_btcs
                    rebal_data_df['previous_ethers'] = previous_ethers
                    rebal_data_df['previous_liquidities'] = previous_liquidities

                    rebal_data_df['new_btcs'] = new_btcs
                    rebal_data_df['new_ethers'] = new_ethers
                    rebal_data_df['new_liquidities'] = new_liquidities

                    rebal_data_df['swapping_sizes'] = swapping_sizes
                    rebal_data_df['swapping_cost'] = swap_costs

                # ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### #######
                # ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### #######
                # ####### ####### ####### ####### end of swapping cost analysis
                # ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### #######
                # ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### #######
                #
                # ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### #######
                # ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### #######
                # ####### ####### ####### ####### incorporation into the backtest
                # ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### #######
                # ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### #######

                stages_df['date'] = pd.to_datetime(stages_df['date'])
                stages_df.set_index('date', inplace=True)
                stages_rebal_df = None
                insert_rebalancing_costs = True
                if insert_rebalancing_costs:
                    stages_rebal_df = pd.merge(stages_df.copy(), rebal_data_df.copy(), suffixes = ['','_rebal'],  how = 'left', right_on='date', left_on='date')
                    stages_rebal_df = stages_rebal_df.set_index('date')
                else :
                    stages_rebal_df = stages_df.copy()

                print('computing the mark to market value')

                def compute_mark_to_market_numeraire_assets(previous_numeraire_size, previous_liquidity_size,previous_price,data_df = None):
                    counter = 0
                    numeraire_sizes = np.zeros(len(data_df))
                    for i, row in data_df.iterrows():
                        rebal = row['rebalancing']
                        if rebal :
                            previous_numeraire_size = row['new_btcs']
                            previous_liquidity_size = row['new_liquidities']
                            previous_price =  row['Close']
                            numeraire_sizes[counter] = previous_numeraire_size
                        else :
                            # we mark to market the asset valo
                            current_price = row['Close']
                            deltanumeraire = get_new_btc_quantity_from_price_change(previous_liquidity_size, current_price, previous_price)
                            numeraire_size = previous_numeraire_size + deltanumeraire
                            numeraire_sizes[counter] = numeraire_size
                        counter = counter + 1
                    return numeraire_sizes
                stages_rebal_df['numeraire_size_mark_to_market'] = compute_mark_to_market_numeraire_assets(to_keep_init_btc, to_keep_init_liq,to_keep_init_price,data_df=stages_rebal_df.copy())

                def compute_mark_to_market_risky_assets(previous_risky_size,previous_liquidity_size,previous_price,data_df = None):
                    counter = 0
                    risky_sizes = np.zeros(len(data_df))
                    for i, row in data_df.iterrows():
                        rebal = row['rebalancing']
                        if rebal :
                            previous_risky_size = row['new_ethers']
                            previous_liquidity_size = row['new_liquidities']
                            previous_price =  row['Close']
                            risky_sizes[counter] = previous_risky_size
                        else :
                            # we mark to market the asset valo
                            current_price = row['Close']
                            delta_risky = get_new_ether_quantity_from_price_change(previous_liquidity_size, current_price, previous_price)
                            risky_size = previous_risky_size + delta_risky
                            risky_sizes[counter] = risky_size
                        counter = counter + 1
                    return risky_sizes

                stages_rebal_df['risky_size_mark_to_market'] = compute_mark_to_market_risky_assets(to_keep_init_eth,to_keep_init_liq,to_keep_init_price,data_df=stages_rebal_df.copy())

                stages_rebal_df['new_btcs'].iloc[0]=to_keep_init_btc
                stages_rebal_df['new_ethers'].iloc[0]=to_keep_init_eth

                stages_rebal_df['new_btcs'] = stages_rebal_df['new_btcs'].ffill()
                stages_rebal_df['new_ethers'] = stages_rebal_df['new_ethers'].ffill()
                stages_rebal_df['new_liquidities'] = stages_rebal_df['new_liquidities'].ffill()

                stages_rebal_df['swapping_sizes'] = stages_rebal_df['swapping_sizes'].fillna(0.)


                stages_rebal_df['new_liquidities'] = stages_rebal_df['new_liquidities'].bfill()
                stages_rebal_df['swapping_sizes'] = stages_rebal_df['swapping_sizes'].bfill()



                stages_rebal_df = stages_rebal_df.rename(columns = {'new_btcs':'numeraire_size',
                                                 'new_ethers':'risky_size',
                                                 'new_liquidities':'liquidity_size'})

                print('investigating the rebalancing datas')


                # ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### #######
                # ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### #######
                # ####### ####### ####### ####### computing the incurred loss
                # ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### #######
                # ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### ####### #######



                all_df = pd.merge(df.copy(),  stages_rebal_df.copy(),suffixes= ['','_rebal'], how = 'left', left_index=True, right_index=True)
                all_df['swapping_cost'] = all_df['swapping_cost'].fillna(0.)




                all_df['apr'] = all_df['apr'].fillna(0.)
                all_df['incurred_v3'] = all_df['incurred_v3'].fillna(0.)
                all_df['incurred_v2'] = all_df['incurred_v2'].fillna(0.)

                all_df['flipped_lp_position_profit'] = all_df['flipped_lp_position_profit'].fillna(0.)
                all_df['lp_position_profit'] = all_df['lp_position_profit'].fillna(0.)
                #all_df[['incurred_v3', 'lp_position_profit', 'we_rebalance_signal', 'preceding_vault_signal','flipped_lp_position_profit']]

                print('computing the gain and losses from all in liguidity')

                all_df['mark_to_market_loss_v2'] = 100. * (1.+all_df['curILv2'].values)



                all_df['loss_v3'] = 100. * np.cumprod(1 + all_df['incurred_v3'].values)
                all_df['loss_v2'] = 100. * np.cumprod(1 + all_df['incurred_v2'].values)

                all_df['curILv3'] = all_df['curILv3'].fillna(0.)
                all_df['rebalancing'] = all_df['rebalancing'].fillna(False)

                def compute_loss_v3_mark_to_market(data_df = None):
                    previous_valo = 100.
                    valos = np.zeros(len(data_df))
                    counter = 0
                    for i, row in data_df.iterrows():
                        rebal = row['rebalancing']
                        current_valo_since_rebal = row['curILv3']
                        valos[counter]= previous_valo * (1. + current_valo_since_rebal)
                        if rebal :
                            previous_valo = valos[counter]
                        counter = counter + 1
                    return valos

                all_df['mark_to_market_loss_v3'] = compute_loss_v3_mark_to_market(data_df=all_df[['incurred_v3', 'curILv3', 'rebalancing']].copy())
                all_df['mark_to_market_loss_v3_returns'] = all_df['mark_to_market_loss_v3'].pct_change()
                all_df['mark_to_market_loss_v3_returns'] = all_df['mark_to_market_loss_v3_returns'].fillna(0.)

                all_df['mark_to_market_loss_v3_without_reset'] = 100. * (1. + all_df['curILv3'].values)


                all_df['cumulated_apr'] =  100. * np.cumprod(1 + all_df['apr'].values)
                all_df['cumulated_swapping_costs'] =  100. * np.cumprod(1  - all_df['swapping_cost'].values)

                all_df['value_LP_LS'] = 100. * np.cumprod(1 + all_df['flipped_lp_position_profit'].values)
                all_df['value_LP_LO'] = 100. * np.cumprod(1 + all_df['lp_position_profit'].values)

                ### computing the mark to market value
                #### curlp_position_profit is the return since the last rebalancing position : the vintage point is reseted each time we rebalance
                ###all_df[['lp_position_profit', 'curlp_position_profit', 'rebalancing','mark_to_market_lp_value', 'value_LP_LO' ]]
                ## this would not work weneed reset
                all_df['mark_to_market_lp_value_without_reset'] = 100. * (1. + all_df['curlp_position_profit'].values)

                def compute_lp_mark_to_market(data_df = None):
                    previous_valo = 100.
                    valos = np.zeros(len(data_df))
                    counter = 0
                    for i, row in data_df.iterrows():
                        rebal = row['rebalancing']
                        current_valo_since_rebal = row['curlp_position_profit']
                        valos[counter]= previous_valo * (1. + current_valo_since_rebal)
                        if rebal :
                            previous_valo = valos[counter]
                        counter = counter + 1
                    return valos

                all_df['curlp_position_profit'] = all_df['curlp_position_profit'].fillna(0.)
                all_df['rebalancing'] = all_df['rebalancing'].fillna(False)

                all_df['mark_to_market_lp_value'] = compute_lp_mark_to_market(data_df = all_df[['lp_position_profit', 'curlp_position_profit', 'rebalancing','value_LP_LO']].copy())
                ###all_df[['lp_position_profit', 'curlp_position_profit', 'rebalancing','mark_to_market_lp_value', 'mark_to_market_lp_value_without_reset', 'value_LP_LO' ]]

                all_df['mark_to_market_lp_value_returns'] = all_df['mark_to_market_lp_value'].pct_change()
                all_df['mark_to_market_lp_value_returns'] = all_df['mark_to_market_lp_value_returns'].fillna(0.)

                all_df['performance_LP_LS'] = 100. * np.cumprod(
                    1 + all_df['flipped_lp_position_profit'].values + all_df['apr'].values - all_df['swapping_cost'].values)
                all_df['performance_LP_LO'] = 100. * np.cumprod(
                    1 + all_df['lp_position_profit'].values + all_df['apr'].values - all_df['swapping_cost'].values)

                all_df['performance_LP_LO_mark_to_market'] = 100. * np.cumprod(1 + all_df['mark_to_market_lp_value_returns'].values + all_df['apr'].values - all_df['swapping_cost'].values)

                all_df['performance_LP_LO'] = 100. * np.cumprod(
                    1 + all_df['lp_position_profit'].values + all_df['apr'].values - all_df['swapping_cost'].values)

                all_df['performance_ILV3_mark_to_market'] = 100. * np.cumprod(1 + all_df['mark_to_market_loss_v3_returns'].values + all_df['apr'].values - all_df['swapping_cost'].values)
                all_df['performance_ILV3'] = 100. * np.cumprod(1 + all_df['incurred_v3'].values + all_df['apr'].values - all_df['swapping_cost'].values)

                #all_df['gain_and_loss_v3'] = 100. * np.cumprod(1 + all_df['incurred_v3'].values + all_df['apr'].values)
                #all_df['gain_and_loss_v2'] = 100. * np.cumprod(1 + all_df['incurred_v2'].values + all_df['apr'].values )

                all_df['performance_against_risky_mark_to_market'] = all_df['performance_LP_LO_mark_to_market']/all_df['Close']
                all_df['performance_against_risky'] = all_df['performance_LP_LO']/all_df['Close']


                if plotHtml:
                    fig = realtime_plotting_utility.plot_multiple_time_series(
                        data_df=all_df[['performance_against_risky_mark_to_market','performance_against_risky']].copy(), logy=False, split=False,
                        put_on_same_scale=False,
                        title=f'Performance against a full risky HODL basket {ssj} {ssj_against}')
                    fig.show()
                    fig = realtime_plotting_utility.plot_multiple_time_series(
                        data_df=all_df[['performance_ILV3_mark_to_market']].copy(), logy=False, split=False,
                        put_on_same_scale=False,
                        title=f'Performance against a 50/50 HODL basket {ssj} {ssj_against}')
                    fig.show()

                    fig = realtime_plotting_utility.plot_multiple_time_series(
                        data_df=all_df[['performance_ILV3_mark_to_market','performance_ILV3']].copy(), logy=False, split=False,
                        put_on_same_scale=False,
                        title=f'Performance against a 50/50 HODL basket {ssj} {ssj_against}')
                    fig.show()
                    fig = realtime_plotting_utility.plot_multiple_time_series(
                        data_df=all_df[['performance_ILV3_mark_to_market']].copy(), logy=False, split=False,
                        put_on_same_scale=False,
                        title=f'Performance against a 50/50 HODL basket {ssj} {ssj_against}')
                    fig.show()
                    fig = realtime_plotting_utility.plot_multiple_time_series(
                        data_df=all_df[['cumulated_swapping_costs']].copy(), logy=False, split=False,
                        put_on_same_scale=False,
                        title=f'swapping costs {ssj} {ssj_against}')
                    fig.show()
                    fig = realtime_plotting_utility.plot_multiple_time_series(
                        data_df=all_df[['swapping_cost']].copy(), logy=False, split=False,
                        put_on_same_scale=False,
                        title=f'swapping costs {ssj} {ssj_against}')
                    fig.show()
                    fig = realtime_plotting_utility.plot_multiple_time_series(
                        data_df=all_df[['swapping_sizes']].copy(), logy=False, split=False,
                        put_on_same_scale=False,
                        title=f'swapping sizes {ssj} {ssj_against}')
                    fig.show()
                    fig = realtime_plotting_utility.plot_multiple_time_series(
                        data_df=all_df[['numeraire_size_mark_to_market','numeraire_size', 'risky_size_mark_to_market', 'risky_size']].copy(), logy=False, split=False,
                        put_on_same_scale=False,
                        title=f'assets {ssj} {ssj_against}')
                    fig.show()
                    fig = realtime_plotting_utility.plot_multiple_time_series(
                        data_df=all_df[['numeraire_size_mark_to_market','numeraire_size']].copy(), logy=False, split=False,
                        put_on_same_scale=False,
                        title=f'numeraire size {ssj} {ssj_against}')
                    fig.show()
                    fig = realtime_plotting_utility.plot_multiple_time_series(
                        data_df=all_df[['risky_size_mark_to_market','risky_size']].copy(), logy=False, split=False,
                        put_on_same_scale=False,
                        title=f'risky {ssj} {ssj_against}')
                    fig.show()
                    fig = realtime_plotting_utility.plot_multiple_time_series(
                        data_df=all_df[['liquidity_size']].copy(), logy=False, split=False,
                        put_on_same_scale=False,
                        title=f'liquidity size {ssj} {ssj_against}')
                    fig.show()

                    fig = realtime_plotting_utility.plot_multiple_time_series(
                        data_df=all_df[['loss_v3', 'mark_to_market_loss_v3']].copy(), logy=False, split=False,
                        put_on_same_scale=False,
                        title=f'impermanent loss V3 {ssj} {ssj_against}')
                    fig.show()

                    fig = realtime_plotting_utility.plot_multiple_time_series(
                        data_df=all_df[['loss_v3', 'mark_to_market_loss_v3']].copy(), logy=False, split=False,
                        put_on_same_scale=False,
                        title=f'loss v3 {ssj} {ssj_against}')
                    fig.show()


                    fig = realtime_plotting_utility.plot_multiple_time_series(
                        data_df=all_df[['loss_v2', 'mark_to_market_loss_v2']].copy(), logy=False, split=False,
                        put_on_same_scale=False,
                        title=f'v3 bounds {ssj} {ssj_against}')
                    fig.show()

                    fig = realtime_plotting_utility.plot_multiple_time_series(
                        data_df=all_df[['value_LP_LO', 'mark_to_market_lp_value']].copy(), logy=False, split=False,
                        put_on_same_scale=False,
                        title=f'mark-to-market/book LP position value {ssj} {ssj_against}')
                    fig.show()

                    fig = realtime_plotting_utility.plot_multiple_time_series(
                        data_df=all_df[['real_lower_bound', 'real_upper_bound', 'Close']].copy(), logy=False, split=False,
                        put_on_same_scale=False,
                        title=f'v3 bounds {ssj} {ssj_against}')
                    fig.show()

                    fig = realtime_plotting_utility.plot_multiple_time_series(
                        data_df=all_df[['signal']].copy(), logy=False, split=False,
                        put_on_same_scale=False,
                        title=f'v3 bounds {ssj} {ssj_against}')
                    fig.show()
                    fig = realtime_plotting_utility.plot_multiple_time_series(
                        data_df=all_df[['curILv3']].copy(), logy=False, split=False,
                        put_on_same_scale=False,
                        title=f'v3 IL {ssj} {ssj_against}')
                    fig.show()
                    fig = realtime_plotting_utility.plot_multiple_time_series(
                        data_df=all_df[['curlp_value_first']].copy(), logy=False, split=False,
                        put_on_same_scale=False,
                        title=f'LP position value {ssj} {ssj_against}')
                    fig.show()
                    fig = realtime_plotting_utility.plot_multiple_time_series(
                        data_df=all_df[['flipped_lp_position_profit']].copy(), logy=False, split=False,
                        put_on_same_scale=False,
                        title=f'LP position value {ssj} {ssj_against}')
                    fig.show()

                    fig1 = realtime_plotting_utility.plot_multiple_time_series(
                        #                    data_df=all_df[['loss_v2', 'loss_v3', 'gain_and_loss_v3','gain_and_loss_LS']].copy(),
                        data_df=all_df[['performance_LP_LS', 'performance_LP_LO']].copy(),
                        logy=False, split=False,
                        put_on_same_scale=False,
                        title=f'Absolute performance LO/LS {ssj} {ssj_against}')
                    fig1.show()

                    fig1 = realtime_plotting_utility.plot_multiple_time_series(
                        #                    data_df=all_df[['loss_v2', 'loss_v3', 'gain_and_loss_v3','gain_and_loss_LS']].copy(),
                        data_df=all_df[['performance_LP_LO_mark_to_market', 'performance_LP_LO']].copy(),
                        logy=False, split=False,
                        put_on_same_scale=False,
                        title=f'Absolute performance v3 liquidity providing {ssj} {ssj_against}')
                    fig1.show()

                    fig1 = realtime_plotting_utility.plot_multiple_time_series(
                        #                    data_df=all_df[['loss_v2', 'loss_v3', 'gain_and_loss_v3','gain_and_loss_LS']].copy(),
                        data_df=all_df[['performance_LP_LO']].copy(),
                        logy=False, split=False,
                        put_on_same_scale=False,
                        title=f'Absolute performance v3 liquidity providing {ssj} {ssj_against}')
                    fig1.show()


                stages_df = all_df.copy()
                # computing the annulized loss
                delta = stages_df.index[len(stages_df) - 1] - stages_df.index[0]
                nb_years = delta.days / 365

                print(f'nb years {nb_years}')
                first_v3_value = stages_df['loss_v3'].iloc[0]
                first_v2_value = stages_df['loss_v2'].iloc[0]

                last_v3_value = stages_df['loss_v3'].iloc[len(stages_df) - 1]
                last_v2_value = stages_df['loss_v2'].iloc[len(stages_df) - 1]

                v2_loss_apy = annualize_percent(first_value=first_v2_value, last_value=last_v2_value,
                                                nb_years=nb_years)
                v3_loss_apy = annualize_percent(first_value=first_v3_value, last_value=last_v3_value,
                                                nb_years=nb_years)

                sum_rebalance = all_df['stage_nb'].max()

                # objective 1 : minimiser le nombre de rebalancements
                obj_rebalance = sum_rebalance
                # objective 2 : minimiser l'impermanent loss
                all_df['providingLiquidity'] = ~np.isnan(all_df['real_upper_bound'])
                on_mode = (all_df['providingLiquidity'] == 1).sum()/len(all_df)
                from cryptotoolbox.risk_metrics import riskmetrics

                kpi_df = riskmetrics.get_kpi(all_df[[f'performance_LP_LO_mark_to_market']])
                obf_data= {
                    'bb_period': bb_period,
                    'stdNbr': stdNbr,
                    'distance_to_bound_perc': distance_to_bound_perc,
                    'mean_pct_change': avg_pct_change,
                    'nb_rebalancing_days': obj_rebalance,
                    'on_mode':on_mode,
                    'v2_loss_apy': v2_loss_apy,
                    'v3_loss_apy': v3_loss_apy,
                }
                obf_data.update(kpi_df.to_dict()['performance_LP_LO_mark_to_market'])

                objectives.append(obf_data)

    print('done')
    objectives_df = pd.DataFrame(objectives)
    return objectives_df, all_df
